/**
   Simple DMA based Linpack benchmark.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "dma-linpack.h"
#include "coherence.h"

#include <tsc.h>
#include <cache.h>
#include <barrier.h>

#include <stdio.h>

/* Temp storage for pivot index and scaling factors. */
static size_t pivot_rows[MATRIX_SIZE];
static double pivot_scales[MATRIX_SIZE];

void 
factorize(double * restrict m,
	  size_t num_equations)
{
  size_t col, max_col, next_panel;
  uint64_t start_t, end_t;
  double flop, time;
  unsigned my_rank;
  uint64_t raw_t;
  
  my_rank = cache_get_mm_id();

  /* We have to also take the B vector into account. */
  max_col = num_equations + 1;

  /* Factorize in panels of columns. */
  for(col=0; col<max_col; col+=PANEL_SIZE) {
  
    start_t = tsc_get();
    
    factorize_panel(pivot_rows + col,
		    pivot_scales + col,
		    m + index_m(col, col),
		    num_equations - col,
		    PANEL_SIZE);

    end_t = tsc_get();

#if 0
    if (my_rank == 0) {
      time = (double)(end_t - start_t);
      flop = num_equations - col;
      flop = 2.0*PANEL_SIZE*PANEL_SIZE*flop - 
	(double)PANEL_SIZE*(PANEL_SIZE + 1)*(2*PANEL_SIZE + 1)/3.0;
      printf("factorize_panel took %llu cycles for %d rows (= %.4f FPC)\n", 
	     end_t - start_t,
	     num_equations - col,
	     flop/time);
    }
#endif

    /* Maybe update the right part of the matrix. */
    next_panel = col + PANEL_SIZE;
    if (next_panel<max_col) {

      coherence_flush_all();
      barrier();

      start_t = tsc_get();
      raw_t = update_m_panel(m + index_m(col, col),
			     pivot_rows + col,
			     PANEL_SIZE,
			     max_col - col);
      coherence_flush_all();
      barrier();
      end_t = tsc_get();

      if (my_rank == 0) {
	time = (double)(end_t - start_t);
	flop = (max_col - col - PANEL_SIZE);
	flop = 2*PANEL_SIZE*flop*flop;

#if 0
	printf("update_m_panel took %llu cycles for %u cols (=%6.4f FPC) "
	       " raw %llu (=%6.4f FPC)\n",
	       end_t - start_t, max_col - col - PANEL_SIZE, 
	       (flop + (double)
		(max_col - col - PANEL_SIZE)*PANEL_SIZE*(PANEL_SIZE + 1))/time,
	       raw_t, 
	       flop/((double)raw_t));
#endif
      }

    }

  }
}
