/**
   Simple DMA based Linpack benchmark.
*/

#include "dma-linpack.h"
#include "coherence.h"

#include <stdio.h>
#include <math.h>
#include <barrier.h>
#include <cache.h>
#include <tsc.h>

/* The panel buffers in MSMC. */
#pragma DATA_SECTION(P_l3, "panel")
double P_l3[2*BLOCK_SIZE*MATRIX_SIZE];

inline size_t 
find_subblock_pivot_row(const double * restrict l, size_t rows);

static void
update_subblock_panel(double * restrict m,
		      const double * restrict l,
		      const size_t * restrict pivot_rows,
		      size_t rows);

static void
factorize_subblock_panel(double * restrict l,
			 size_t * restrict pivot_rows,
			 double * restrict pivot_scales,
			 size_t rows,
			 size_t row_offset);

static void
update_subblock_panel(double * restrict m,
		      const double * restrict l,
		      const size_t * restrict pivot_rows,
		      size_t rows);

static void
load_panel_block(double * restrict l_l3,
		 const double * restrict l,
		 size_t rows);

static void
save_panel_block(double * restrict l,
		 const double * restrict l_l3,
		 size_t rows);

static void
update_panel_block_u(double * restrict m,
		     const double * restrict l_l3,
		     const size_t * restrict pivot_rows,
		     size_t pivot_ofs,
		     size_t panel_width);

static void
update_panel_block_m(double * restrict m,
		     double * restrict l_l3_next,
		     const double * restrict l_l3,
		     const double * restrict u,
		     size_t panel_width,
		     size_t rows,
		     size_t row_stride);

void
factorize_panel(size_t * restrict pivot_rows,
		double * restrict pivot_scales,
		double * restrict m,
		size_t rows,
		size_t panel_width)
{
  size_t col_block;
  size_t col_ofs;
  size_t upd_block;
  size_t upd_ofs;
  size_t col;
  size_t pivot_row;
  size_t row;
  double tmp;
  double * restrict trp;
  double * restrict prp;
  unsigned my_rank;
  unsigned num_ranks;
  size_t row_stride;
  size_t row_start;
  double * restrict l_cur;
  double * restrict l_next;
  double * restrict l_tmp;
#if 0
  uint64_t start_t, end_t;
  uint64_t accum_l, accum_u, accum_m, accum_c;

  accum_l = 0;
  accum_u = 0;
  accum_m = 0;
  accum_c = 0;
#endif

  my_rank = cache_get_mm_id();
  num_ranks = cache_get_mm_num();
  
  l_cur = P_l3;
  l_next = P_l3 + MATRIX_SIZE*BLOCK_SIZE;

  if (my_rank == 0) {
#if 0
    accum_c -= tsc_get();
#endif
    load_panel_block(l_cur, m, rows);
#if 0
    accum_c += tsc_get();
#endif
  }

  for(col_block=0; col_block<panel_width; col_block+=BLOCK_SIZE) {
#if 0
    start_t = tsc_get();
#endif
    if (my_rank == 0) {

      /* Factorize the block panel. */
      for(col_ofs=0; col_ofs<BLOCK_SIZE; col_ofs+=SUBBLOCK_SIZE) {
	factorize_subblock_panel(l_cur + col_ofs*(1 + MATRIX_SIZE),
				 pivot_rows + col_block + col_ofs,
				 pivot_scales + col_block + col_ofs,
				 rows - col_block - col_ofs,
				 0);

	for(upd_ofs=col_ofs+SUBBLOCK_SIZE;
	    upd_ofs<BLOCK_SIZE; 
	    upd_ofs+=SUBBLOCK_SIZE) {
	  update_subblock_panel(l_cur + col_ofs + upd_ofs*MATRIX_SIZE,
				l_cur + col_ofs + col_ofs*MATRIX_SIZE,
				pivot_rows + col_block + col_ofs,
				rows - col_block - col_ofs);
	}

      }

      /* Pivot left part of rows in the panel. */
      for(col_ofs=0; col_ofs<BLOCK_SIZE; col_ofs+=SUBBLOCK_SIZE) {
	for(upd_ofs=0; upd_ofs<SUBBLOCK_SIZE; ++upd_ofs) {
	  pivot_row = pivot_rows[col_block + col_ofs + upd_ofs];
	  for(col=0; col<col_ofs; ++col) {
	    tmp = l_cur[upd_ofs + col_ofs + col*MATRIX_SIZE];
	    l_cur[upd_ofs + col_ofs + col*MATRIX_SIZE] = 
	      l_cur[pivot_row + col_ofs + col*MATRIX_SIZE];
	    l_cur[pivot_row + col_ofs + col*MATRIX_SIZE] = tmp;
	  }
	}
      }

#if 0
      accum_c -= tsc_get();
#endif
      save_panel_block(m + index_m(col_block, col_block), 
		       l_cur, 
		       rows - col_block);
#if 0
      accum_c += tsc_get();
#endif

      /* Adjust pivot_row offsets. */
      for(col_ofs=0; col_ofs<BLOCK_SIZE; ++col_ofs) 
	pivot_rows[col_block + col_ofs] += 
	  (col_ofs/SUBBLOCK_SIZE)*SUBBLOCK_SIZE + col_block;

      /* Pivot left blocks in the panel. */
      if (col_block >= BLOCK_SIZE) {
	for(row=0; row<BLOCK_SIZE; ++row) {
	  pivot_row = pivot_rows[col_block + row];

	  for(upd_block=0; upd_block<col_block; upd_block+=BLOCK_SIZE) {
	    trp = m + index_m(col_block + row, upd_block);
	    prp = m + index_m(pivot_row, upd_block);
	  
	    for(col=0; col<BLOCK_SIZE; ++col) {
	      tmp = trp[col];
	      trp[col] = prp[col];
	      prp[col] = tmp;
	    }
	  }
	}
      }
    }
    
#if 0
    end_t = tsc_get();
    accum_l += end_t - start_t;
#endif

    /* Update the trailing panels. */
    upd_block = col_block + BLOCK_SIZE;
    if (upd_block < panel_width) {

#if 0
      start_t = tsc_get();
#endif
      
      row_start = upd_block + BLOCK_SIZE*my_rank;
      row_stride = BLOCK_SIZE*num_ranks;

      if (my_rank == 0) {
	update_panel_block_u(m + index_m(col_block, upd_block),
			     l_cur,
			     pivot_rows + col_block,
			     col_block,
			     panel_width - upd_block);
      }
   
      coherence_flush_all();
      barrier();

#if 0
      end_t = tsc_get();
      accum_u += end_t - start_t;
 
      start_t = tsc_get();
#endif


      if (row_start < rows) {
	update_panel_block_m(m + index_m(row_start, upd_block),
			     l_next + BLOCK_SIZE*my_rank,
			     l_cur + BLOCK_SIZE*(my_rank + 1),
			     m + index_m(col_block, upd_block),
			     panel_width - upd_block,
			     rows - row_start + 1,
			     row_stride);
      }

      coherence_flush_all();
      barrier();
#if 0
      end_t = tsc_get();
      accum_m += end_t - start_t;
#endif

    } 

    l_tmp = l_cur;
    l_cur = l_next; 
    l_next = l_tmp;
  }

  coherence_flush_all();
  barrier();

#if 0
  if (my_rank==0) {
    printf("factorize_panel l %llu  u %llu  m %llu  c %llu!\n",
	   accum_l, accum_u, accum_m, accum_c);
  }
#endif
}
		
inline size_t
find_subblock_pivot_row(const double * restrict l, size_t rows)
{
  size_t row;
  size_t pivot_row;
  double pivot_abs, cur_abs;

  pivot_row = 0;
  pivot_abs = fabs(l[0]);
  
  for(row = 1; row<rows; ++row) {
    cur_abs = fabs(l[row]);
    if (cur_abs > pivot_abs) {
      pivot_row = row;
      pivot_abs = cur_abs;
    }
  }

  return pivot_row;
}

void
factorize_subblock_panel(double * restrict l,
			 size_t * restrict pivot_rows,
			 double * restrict pivot_scales,
			 size_t rows,
			 size_t row_offset)
{
  

  size_t row;
  size_t pivot_row;
  double pivot_abs;
  double cur_abs;
  double * restrict l_cur;
  double pivot_scale;

  double l_0, l_1, l_2, l_3;
  double u_01, u_02, u_03;
  double u_12, u_13;
  double u_23;

  /* Column 0 */

  /* Find the pivot for the column. */
  pivot_row = 0;
  pivot_abs = fabs(l[0]);
  for(l_cur=l+1, row=1; row<rows; ++l_cur, ++row) {
    cur_abs = fabs(*l_cur);
    if (cur_abs > pivot_abs) {
      pivot_row = row;
      pivot_abs = cur_abs;
    }
  }
  
  /* Scale and Swap the pivot row with row 0. */
  l_0 = l[pivot_row ];
  l[pivot_row] = l[0];
  pivot_scale=1.0/l_0;
  l[0] = pivot_scale;

  u_01 = l[pivot_row +   MATRIX_SIZE];
  l[pivot_row + MATRIX_SIZE] = l[MATRIX_SIZE];
  u_01 *= pivot_scale;
  l[MATRIX_SIZE] = u_01;

  u_02 = l[pivot_row + 2*MATRIX_SIZE];
  l[pivot_row + 2*MATRIX_SIZE] = l[2*MATRIX_SIZE];
  u_02 *= pivot_scale;
  l[2*MATRIX_SIZE] = u_02;

  u_03 = l[pivot_row + 3*MATRIX_SIZE];
  l[pivot_row + 3*MATRIX_SIZE] = l[3*MATRIX_SIZE];
  u_03 *= pivot_scale;
  l[3*MATRIX_SIZE] = u_03;

  pivot_rows[0] = pivot_row;
  pivot_scales[0] = pivot_scale;

  
  /* Column 1 */

  /* Update the column & find new pivot row. */
  pivot_row = 1;
  l_0 = l[1];
  pivot_scale = l[1 + MATRIX_SIZE];
  pivot_scale = pivot_scale - l_0*u_01;
  l[1 + MATRIX_SIZE] = pivot_scale; /* CHECK (redundant?) */
  pivot_abs = fabs(pivot_scale);

  for(row=2; row<rows; ++row) {
    l_0 = l[row];
    l_1 = l[row + MATRIX_SIZE];
    l_1 = l_1 - l_0*u_01;
    l[row + MATRIX_SIZE] = l_1;

    cur_abs = fabs(l_1);
    if (cur_abs > pivot_abs) {
      pivot_row = row;
      pivot_abs = cur_abs;
    }
  }

  /* Scale and Swap the pivot row with row 1. */
  l_0 = l[pivot_row ];
  l[pivot_row] = l[1];
  l[1] = l_0;

  l_1 = l[pivot_row + MATRIX_SIZE];
  l[pivot_row + MATRIX_SIZE] = l[1 + MATRIX_SIZE];
  pivot_scale = 1.0/l_1;
  l[1 + MATRIX_SIZE] = pivot_scale;

  u_12 = l[pivot_row + 2*MATRIX_SIZE];
  l[pivot_row + 2*MATRIX_SIZE] = l[1 + 2*MATRIX_SIZE];
  u_12 = pivot_scale*(u_12 - l_0*u_02);
  l[1 + 2*MATRIX_SIZE] = u_12;

  u_13 = l[pivot_row + 3*MATRIX_SIZE];
  l[pivot_row + 3*MATRIX_SIZE] = l[1 + 3*MATRIX_SIZE];
  u_13 = pivot_scale*(u_13 - l_0*u_03);
  l[1 + 3*MATRIX_SIZE] = u_13;

  pivot_rows[1] = pivot_row;
  pivot_scales[1] = pivot_scale;

  /* Column 2 */

  /* Update the column & find new pivot row. */
  pivot_row = 2;
  l_0 = l[2];
  l_1 = l[2 + MATRIX_SIZE];
  pivot_scale = l[2 + 2*MATRIX_SIZE];
  pivot_scale = pivot_scale - l_0*u_02 - l_1*u_12;
  l[2 + 2*MATRIX_SIZE] = pivot_scale; /* CHECK (redundant?) */
  pivot_abs = fabs(pivot_scale);

  for(row=3; row<rows; ++row) {
    l_0 = l[row];
    l_1 = l[row + MATRIX_SIZE];
    l_2 = l[row + 2*MATRIX_SIZE];
    l_2 = l_2 - l_0*u_02 - l_1*u_12;
    l[row + 2*MATRIX_SIZE] = l_2;

    cur_abs = fabs(l_2);
    if (cur_abs > pivot_abs) {
      pivot_row = row;
      pivot_abs = cur_abs;
    }
  }

  /* Scale and Swap the pivot row with row 2. */
  l_0 = l[pivot_row];
  l[pivot_row] = l[2];
  l[2] = l_0;

  l_1 = l[pivot_row + MATRIX_SIZE];
  l[pivot_row + MATRIX_SIZE] = l[2 + MATRIX_SIZE];
  l[2 + MATRIX_SIZE] = l_1;

  l_2 = l[pivot_row + 2*MATRIX_SIZE];
  l[pivot_row + 2*MATRIX_SIZE] = l[2 + 2*MATRIX_SIZE];
  pivot_scale = 1.0/l_2;
  l[2 + 2*MATRIX_SIZE] = pivot_scale;

  u_23 = l[pivot_row + 3*MATRIX_SIZE];
  l[pivot_row + 3*MATRIX_SIZE] = l[2 + 3*MATRIX_SIZE];
  u_23 = pivot_scale*(u_23 - l_0*u_03 - l_1*u_13);
  l[2 + 3*MATRIX_SIZE] = u_23;

  pivot_rows[2] = pivot_row;
  pivot_scales[2] = pivot_scale;

  /* Column 3 */

  /* Update the column & find new pivot row. */
  pivot_row = 3;
  l_0 = l[3];
  l_1 = l[3 + MATRIX_SIZE];
  l_2 = l[3 + 2*MATRIX_SIZE];
  
  pivot_scale = l[3 + 3*MATRIX_SIZE];
  pivot_scale = pivot_scale - l_0*u_03 - l_1*u_13 - l_2*u_23;
  l[3 + 3*MATRIX_SIZE] = pivot_scale; /* CHECK (redundant?) */
  pivot_abs = fabs(pivot_scale);

  for(row=4; row<rows; ++row) {
    l_0 = l[row];
    l_1 = l[row + MATRIX_SIZE];
    l_2 = l[row + 2*MATRIX_SIZE];
    l_3 = l[row + 3*MATRIX_SIZE];
    l_3 = l_3 - l_0*u_03 - l_1*u_13 - l_2*u_23;
    l[row + 3*MATRIX_SIZE] = l_3;

    cur_abs = fabs(l_3);
    if (cur_abs > pivot_abs) {
      pivot_row = row;
      pivot_abs = cur_abs;
    }
  }

  /* Scale and Swap the pivot row with row 3. */
  l_0 = l[pivot_row];
  l[pivot_row] = l[3];
  l[3] = l_0;

  l_1 = l[pivot_row + MATRIX_SIZE];
  l[pivot_row + MATRIX_SIZE] = l[3 + MATRIX_SIZE];
  l[3 + MATRIX_SIZE] = l_1;

  l_2 = l[pivot_row + 2*MATRIX_SIZE];
  l[pivot_row + 2*MATRIX_SIZE] = l[3 + 2*MATRIX_SIZE];
  l[3 + 2*MATRIX_SIZE] = l_2;

  l_3 = l[pivot_row + 3*MATRIX_SIZE];
  l[pivot_row + 3*MATRIX_SIZE] = l[3 + 3*MATRIX_SIZE];
  pivot_scale = 1.0/l_3;
  l[3 + 3*MATRIX_SIZE] = pivot_scale;

  pivot_rows[3] = pivot_row;
  pivot_scales[3] = pivot_scale;
}

void
update_subblock_panel(double * restrict m,
		      const double * restrict l,
		      const size_t * restrict pivot_rows,
		      size_t rows)
{
  double l_0, l_1, l_2, l_3;
  
  double u_00, u_01, u_02, u_03;
  double u_10, u_11, u_12, u_13;
  double u_20, u_21, u_22, u_23;
  double u_30, u_31, u_32, u_33;

  double m_0, m_1, m_2, m_3;

  size_t row;
 
  double * restrict mp;

  /* Exchange and scale the pivot rows. */

  l_0 = l[0];
  mp = m + pivot_rows[0];
  u_00 = l_0*mp[0];
  mp[0] = m[0];
  m[0] = u_00;

  u_01 = l_0*mp[MATRIX_SIZE];
  mp[MATRIX_SIZE] = m[MATRIX_SIZE];
  m[MATRIX_SIZE] = u_01;

  u_02 = l_0*mp[2*MATRIX_SIZE];
  mp[2*MATRIX_SIZE] = m[2*MATRIX_SIZE];
  m[2*MATRIX_SIZE] = u_02;

  u_03 = l_0*mp[3*MATRIX_SIZE];
  mp[3*MATRIX_SIZE] = m[3*MATRIX_SIZE];
  m[3*MATRIX_SIZE] = u_03;

  l_0 = l[1];
  l_1 = l[1 + MATRIX_SIZE];
  mp = m + pivot_rows[1];
  u_10 = l_1*(mp[0] - l_0*u_00);
  mp[0] = m[1];
  m[1] = u_10;

  u_11 = l_1*(mp[MATRIX_SIZE] - l_0*u_01);
  mp[MATRIX_SIZE] = m[1 + MATRIX_SIZE];
  m[1 + MATRIX_SIZE] = u_11;

  u_12 = l_1*(mp[2*MATRIX_SIZE] - l_0*u_02);
  mp[2*MATRIX_SIZE] = m[1 + 2*MATRIX_SIZE];
  m[1 + 2*MATRIX_SIZE] = u_12;

  u_13 = l_1*(mp[3*MATRIX_SIZE] - l_0*u_03);
  mp[3*MATRIX_SIZE] = m[1 + 3*MATRIX_SIZE];
  m[1 + 3*MATRIX_SIZE] = u_13;

  l_0 = l[2];
  l_1 = l[2 + MATRIX_SIZE];
  l_2 = l[2 + 2*MATRIX_SIZE];
  mp = m + pivot_rows[2];
  u_20 = l_2*(mp[0] - l_0*u_00 - l_1*u_10);
  mp[0] = m[2];
  m[2] = u_20;

  u_21 = l_2*(mp[MATRIX_SIZE] - l_0*u_01 - l_1*u_11);
  mp[MATRIX_SIZE] = m[2 + MATRIX_SIZE];
  m[2 + MATRIX_SIZE] = u_21;

  u_22 = l_2*(mp[2*MATRIX_SIZE] - l_0*u_02 - l_1*u_12);
  mp[2*MATRIX_SIZE] = m[2 + 2*MATRIX_SIZE];
  m[2 + 2*MATRIX_SIZE] = u_22;

  u_23 = l_2*(mp[3*MATRIX_SIZE] - l_0*u_03 - l_1*u_13);
  mp[3*MATRIX_SIZE] = m[2 + 3*MATRIX_SIZE];
  m[2 + 3*MATRIX_SIZE] = u_23;

  l_0 = l[3];
  l_1 = l[3 + MATRIX_SIZE];
  l_2 = l[3 + 2*MATRIX_SIZE];
  l_3 = l[3 + 3*MATRIX_SIZE];
  mp = m + pivot_rows[3];
  u_30 = l_3*(mp[0] - l_0*u_00 - l_1*u_10 - l_2*u_20);
  mp[0] = m[3];
  m[3] = u_30;

  u_31 = l_3*(mp[MATRIX_SIZE] - l_0*u_01 - l_1*u_11 - l_2*u_21);
  mp[MATRIX_SIZE] = m[3 + MATRIX_SIZE];
  m[3 + MATRIX_SIZE] = u_31;

  u_32 = l_3*(mp[2*MATRIX_SIZE] - l_0*u_02 - l_1*u_12 - l_2*u_22);
  mp[2*MATRIX_SIZE] = m[3 + 2*MATRIX_SIZE];
  m[3 + 2*MATRIX_SIZE] = u_32;

  u_33 = l_3*(mp[3*MATRIX_SIZE] - l_0*u_03 - l_1*u_13 - l_2*u_23);
  mp[3*MATRIX_SIZE] = m[3 + 3*MATRIX_SIZE];
  m[3 + 3*MATRIX_SIZE] = u_33;

  /* Update the lower part of the panel. */
  for(row = 4; row<rows; ++row) {
    l_0 = l[row];
    l_1 = l[row + MATRIX_SIZE];
    l_2 = l[row + 2*MATRIX_SIZE];
    l_3 = l[row + 3*MATRIX_SIZE];

    m_0 = m[row];
    m_1 = m[row + MATRIX_SIZE];
    m_2 = m[row + 2*MATRIX_SIZE];
    m_3 = m[row + 3*MATRIX_SIZE];

    m_0 = m_0 - l_0*u_00 - l_1*u_10 - l_2*u_20 - l_3*u_30;
    m_1 = m_1 - l_0*u_01 - l_1*u_11 - l_2*u_21 - l_3*u_31;
    m_2 = m_2 - l_0*u_02 - l_1*u_12 - l_2*u_22 - l_3*u_32;
    m_3 = m_3 - l_0*u_03 - l_1*u_13 - l_2*u_23 - l_3*u_33;

    m[row] = m_0;
    m[row + MATRIX_SIZE] = m_1;
    m[row + 2*MATRIX_SIZE] = m_2;
    m[row + 3*MATRIX_SIZE] = m_3;
  }
}

void
load_panel_block(double * restrict l_l3,
		 const double * restrict l,
		 size_t rows)
{
  size_t row_block;
  size_t row_ofs, col;

  const double * restrict lp;
  
  for(row_block=0; row_block<rows; row_block+=BLOCK_SIZE) {
    lp = l + row_block*MATRIX_SIZE;

    for(row_ofs=0; row_ofs<BLOCK_SIZE; ++row_ofs) {
      for(col=0; col<BLOCK_SIZE; ++col) {
	l_l3[row_block + row_ofs + col*MATRIX_SIZE] =
	  lp[row_ofs*BLOCK_SIZE + col];
      }
    }
  }
}

void
save_panel_block(double * restrict l,
		 const double * restrict l_l3,
		 size_t rows)
{
  size_t row_block;
  size_t row_ofs, col;

  double * restrict lp;
  
  for(row_block=0; row_block<rows; row_block+=BLOCK_SIZE) {
    lp = l + row_block*MATRIX_SIZE;

    for(row_ofs=0; row_ofs<BLOCK_SIZE; ++row_ofs) {
      for(col=0; col<BLOCK_SIZE; ++col) {
	lp[row_ofs*BLOCK_SIZE + col] = 	
	  l_l3[row_block + row_ofs + col*MATRIX_SIZE];
      }
    }
  }
}


void
update_panel_block_u(double * restrict m,
		     const double * restrict l_l3,
		     const size_t * restrict pivot_rows,
		     size_t pivot_ofs,
		     size_t panel_width)
{
  size_t col_block;
  size_t col_ofs;
  size_t row;
  size_t pivot_row;
  size_t r;
  double tmp;
  double * restrict trp;
  double * restrict prp;

  double sum_00, sum_01, sum_02, sum_03;
  double sum_10, sum_11, sum_12, sum_13;
  double sum_20, sum_21, sum_22, sum_23;
  double sum_30, sum_31, sum_32, sum_33;

  double l_0, l_1, l_2, l_3;
  double u_0, u_1, u_2, u_3;

  double * restrict u;
  const double * restrict l;

  for(col_block=0; col_block<panel_width; col_block+=BLOCK_SIZE) {
    for(row=0; row<BLOCK_SIZE; ++row) {
      pivot_row = pivot_rows[row] - pivot_ofs;
      if (pivot_row == row) continue;

      prp = m + index_m(pivot_row, col_block);
      trp = m + index_m(row, col_block);
      
      for(col_ofs=0; col_ofs<BLOCK_SIZE; ++col_ofs) {
	tmp = prp[col_ofs];
	prp[col_ofs] = trp[col_ofs];
	trp[col_ofs] = tmp;
      } 
    }
  }

  for(col_block=0; col_block<panel_width; col_block+=BLOCK_SIZE) {
    for(row=0; row<BLOCK_SIZE; row+=SUBBLOCK_SIZE) {
      for(col_ofs=0; col_ofs<BLOCK_SIZE; col_ofs += SUBBLOCK_SIZE) {

	u = m + index_m(0, col_ofs + col_block);
	l = l_l3 + row;

	sum_00 = 0.0; sum_01 = 0.0; sum_02 = 0.0; sum_03 = 0.0;
	sum_10 = 0.0; sum_11 = 0.0; sum_12 = 0.0; sum_13 = 0.0;
	sum_20 = 0.0; sum_21 = 0.0; sum_22 = 0.0; sum_23 = 0.0;
	sum_30 = 0.0; sum_31 = 0.0; sum_32 = 0.0; sum_33 = 0.0;

	for(r=row; r>0; --r) {
	  u_0 = u[0]; u_1 = u[1]; u_2 = u[2]; u_3 = u[3];
	  l_0 = l[0]; l_1 = l[1]; l_2 = l[2]; l_3 = l[3];
	  u+=BLOCK_SIZE; 
	  l+=MATRIX_SIZE;

	  sum_00 += u_0*l_0;
	  sum_01 += u_1*l_0;
	  sum_02 += u_2*l_0;
	  sum_03 += u_3*l_0;

	  sum_10 += u_0*l_1;
	  sum_11 += u_1*l_1;
	  sum_12 += u_2*l_1;
	  sum_13 += u_3*l_1;

	  sum_20 += u_0*l_2;
	  sum_21 += u_1*l_2;
	  sum_22 += u_2*l_2;
	  sum_23 += u_3*l_2;

	  sum_30 += u_0*l_3;
	  sum_31 += u_1*l_3;
	  sum_32 += u_2*l_3;
	  sum_33 += u_3*l_3;
	}

	u_0 = u[0]; u_1 = u[1]; u_2 = u[2]; u_3 = u[3];
	l_0 = l[0]; l_1 = l[1]; l_2 = l[2]; l_3 = l[3];

	sum_00 = l_0*(u_0 - sum_00); 
	sum_01 = l_0*(u_1 - sum_01);
	sum_02 = l_0*(u_2 - sum_02);
	sum_03 = l_0*(u_3 - sum_03);

	u[0] = sum_00;
	u[1] = sum_01;
	u[2] = sum_02;
	u[3] = sum_03;

	u+=BLOCK_SIZE;
	l+=MATRIX_SIZE;

	sum_10 += sum_00*l_1;
	sum_11 += sum_01*l_1;
	sum_12 += sum_02*l_1;
	sum_13 += sum_03*l_1;
	
	sum_20 += sum_00*l_2;
	sum_21 += sum_01*l_2;
	sum_22 += sum_02*l_2;
	sum_23 += sum_03*l_2;

	sum_30 += sum_00*l_3;
	sum_31 += sum_01*l_3;
	sum_32 += sum_02*l_3;
	sum_33 += sum_03*l_3;

	u_0 = u[0]; u_1 = u[1]; u_2 = u[2]; u_3 = u[3];
	l_1 = l[1]; l_2 = l[2]; l_3 = l[3];

	sum_10 = l_1*(u_0 - sum_10); 
	sum_11 = l_1*(u_1 - sum_11);
	sum_12 = l_1*(u_2 - sum_12);
	sum_13 = l_1*(u_3 - sum_13);

	u[0] = sum_10;
	u[1] = sum_11;
	u[2] = sum_12;
	u[3] = sum_13;

	u+=BLOCK_SIZE;
	l+=MATRIX_SIZE;
	
	sum_20 += sum_10*l_2;
	sum_21 += sum_11*l_2;
	sum_22 += sum_12*l_2;
	sum_23 += sum_13*l_2;

	sum_30 += sum_10*l_3;
	sum_31 += sum_11*l_3;
	sum_32 += sum_12*l_3;
	sum_33 += sum_13*l_3;

	u_0 = u[0]; u_1 = u[1]; u_2 = u[2]; u_3 = u[3];
	l_2 = l[2]; l_3 = l[3];

	sum_20 = l_2*(u_0 - sum_20); 
	sum_21 = l_2*(u_1 - sum_21);
	sum_22 = l_2*(u_2 - sum_22);
	sum_23 = l_2*(u_3 - sum_23);

	u[0] = sum_20;
	u[1] = sum_21;
	u[2] = sum_22;
	u[3] = sum_23;

	u+=BLOCK_SIZE;
	l+=MATRIX_SIZE;
	
	sum_30 += sum_20*l_3;
	sum_31 += sum_21*l_3;
	sum_32 += sum_22*l_3;
	sum_33 += sum_23*l_3;

	u_0 = u[0]; u_1 = u[1]; u_2 = u[2]; u_3 = u[3];
	l_3 = l[3];

	sum_30 = l_3*(u_0 - sum_30); 
	sum_31 = l_3*(u_1 - sum_31);
	sum_32 = l_3*(u_2 - sum_32);
	sum_33 = l_3*(u_3 - sum_33);

	u[0] = sum_30;
	u[1] = sum_31;
	u[2] = sum_32;
	u[3] = sum_33;	
      }
    }
  }
}

void
update_panel_block_m(double * restrict m,
		     double * restrict l_l3_next,
		     const double * restrict l_l3,
		     const double * restrict u,
		     size_t panel_width,
		     size_t rows,
		     size_t row_stride)
{
  size_t row_block;
  size_t row_ofs;
  size_t col_block;
  size_t col_ofs;
  size_t k;
  
  const double * restrict up;
  const double * restrict lp;
  double * restrict mp;
  double * restrict ln_p;

  double ll0, ll1, ll2, ll3;
  double ul0, ul1, ul2, ul3;

  double ml00, ml01, ml02, ml03;
  double ml10, ml11, ml12, ml13;
  double ml20, ml21, ml22, ml23;
  double ml30, ml31, ml32, ml33;

  for(row_block=0; row_block<rows; row_block+=row_stride) {
    for(col_block=0; col_block<panel_width; col_block+=BLOCK_SIZE) {
      up = u + col_block*BLOCK_SIZE;
      mp = m + row_block*MATRIX_SIZE + col_block*BLOCK_SIZE;

      for(row_ofs=0; row_ofs<BLOCK_SIZE; row_ofs+=SUBBLOCK_SIZE) {
	lp = l_l3 + row_block + row_ofs;
	for(col_ofs=0; col_ofs<BLOCK_SIZE; col_ofs+=SUBBLOCK_SIZE) {
	  ul0 = *(up++);
	  ul1 = *(up++);
	  ul2 = *(up++);
	  ul3 = *(up); 
	  up += BLOCK_SIZE - 3;
      
	  ll0 = *(lp++);
	  ll1 = *(lp++);
	  ll2 = *(lp++);
	  ll3 = *(lp); 
	  lp += MATRIX_SIZE - 3;
      
	  ml00 = ll0*ul0;
	  ml01 = ll0*ul1;
	  ml02 = ll0*ul2;
	  ml03 = ll0*ul3;
      
	  ml10 = ll1*ul0;
	  ml11 = ll1*ul1;
	  ml12 = ll1*ul2;
	  ml13 = ll1*ul3;
      
	  ml20 = ll2*ul0;
	  ml21 = ll2*ul1;
	  ml22 = ll2*ul2;
	  ml23 = ll2*ul3;
      
	  ml30 = ll3*ul0;
	  ml31 = ll3*ul1;
	  ml32 = ll3*ul2;
	  ml33 = ll3*ul3;

	  for(k=BLOCK_SIZE - 1; k>0; --k) {
	    ul0 = *(up++);
	    ul1 = *(up++);
	    ul2 = *(up++);
	    ul3 = *(up); 
	    up += BLOCK_SIZE - 3;

	    ll0 = *(lp++);
	    ll1 = *(lp++);
	    ll2 = *(lp++);
	    ll3 = *(lp); 
	    lp += MATRIX_SIZE - 3;

	    ml00 += ll0*ul0;
	    ml01 += ll0*ul1;
	    ml02 += ll0*ul2;
	    ml03 += ll0*ul3;

	    ml10 += ll1*ul0;
	    ml11 += ll1*ul1;
	    ml12 += ll1*ul2;
	    ml13 += ll1*ul3;

	    ml20 += ll2*ul0;
	    ml21 += ll2*ul1;
	    ml22 += ll2*ul2;
	    ml23 += ll2*ul3;

	    ml30 += ll3*ul0;
	    ml31 += ll3*ul1;
	    ml32 += ll3*ul2;
	    ml33 += ll3*ul3;
	  } /* k */

	  if (col_block == 0) {
	    ln_p = l_l3_next + col_ofs*MATRIX_SIZE + row_block + row_ofs;
	    
	    ln_p[            0] = *(mp++) - ml00;
	    ln_p[  MATRIX_SIZE] = *(mp++) - ml01;
	    ln_p[2*MATRIX_SIZE] = *(mp++) - ml02;
	    ln_p[3*MATRIX_SIZE] = *(mp) - ml03;
	    mp += BLOCK_SIZE - 3;

	    ln_p[1                ] = *(mp++) - ml10;
	    ln_p[1 +   MATRIX_SIZE] = *(mp++) - ml11;
	    ln_p[1 + 2*MATRIX_SIZE] = *(mp++) - ml12;
	    ln_p[1 + 3*MATRIX_SIZE] = *(mp) - ml13;
	    mp += BLOCK_SIZE - 3;

	    ln_p[2                ] = *(mp++) - ml20;
	    ln_p[2 +   MATRIX_SIZE] = *(mp++) - ml21;
	    ln_p[2 + 2*MATRIX_SIZE] = *(mp++) - ml22;
	    ln_p[2 + 3*MATRIX_SIZE] = *(mp) - ml23;
	    mp += BLOCK_SIZE - 3;

	    ln_p[3                ] = *(mp++) - ml30;
	    ln_p[3 +   MATRIX_SIZE] = *(mp++) - ml31;
	    ln_p[3 + 2*MATRIX_SIZE] = *(mp++) - ml32;
	    ln_p[3 + 3*MATRIX_SIZE] = *(mp) - ml33;
	    mp -= 3*BLOCK_SIZE - 1;
	  } else {
	    *(mp++) = *(mp) - ml00;
	    *(mp++) = *(mp) - ml01;
	    *(mp++) = *(mp) - ml02;
	    *(mp) = *(mp) - ml03;
	    mp += BLOCK_SIZE - 3;

	    *(mp++) = *(mp) - ml10;
	    *(mp++) = *(mp) - ml11;
	    *(mp++) = *(mp) - ml12;
	    *(mp) = *(mp) - ml13;
	    mp += BLOCK_SIZE - 3;

	    *(mp++) = *(mp) - ml20;
	    *(mp++) = *(mp) - ml21;
	    *(mp++) = *(mp) - ml22;
	    *(mp) = *(mp) - ml23;
	    mp += BLOCK_SIZE - 3;

	    *(mp++) = *(mp) - ml30;
	    *(mp++) = *(mp) - ml31;
	    *(mp++) = *(mp) - ml32;
	    *(mp) = *(mp) - ml33;
	    mp -= 3*BLOCK_SIZE - 1;
	  }

	  lp -= BLOCK_SIZE*MATRIX_SIZE;
	  up -= BLOCK_SIZE*BLOCK_SIZE - 4;
	} /* col_ofs */
	
	up -= BLOCK_SIZE;
	mp += (SUBBLOCK_SIZE - 1)*BLOCK_SIZE;
      } /* row_ofs */
    } /* col_block */
  } /* row_block */
}
