/**
   Simple DMA based Linpack benchmark.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "dma-linpack.h"

void 
fill_u_panel_l2(double * restrict u_l2,
		double * restrict m,
		const size_t * restrict pivot_rows,
		size_t rows,
		size_t cols)
{
  size_t col_block;
  size_t col_ofs;
  size_t row;
  size_t pivot_row;

  for(row=0; row<rows; ++row) {
    pivot_row = pivot_rows[row];

    for(col_block=0; col_block<cols; col_block+=BLOCK_SIZE*8) {
      
      for(col_ofs=0; col_ofs<BLOCK_SIZE; ++col_ofs)
	u_l2[index_u_l2(row, col_block/8 + col_ofs)] = 
	  m[index_m(pivot_row, col_block + col_ofs)];
    
      if (pivot_row != row) {
	for(col_ofs=0; col_ofs<BLOCK_SIZE; ++col_ofs)
	  m[index_m(pivot_row, col_block + col_ofs)] =
	    m[index_m(row, col_block + col_ofs)];
      }
    }
  }
}

void
fill_l_head_l2(double * restrict l_head,
	       const double * restrict m,
	       size_t rows,
	       size_t cols)
{
  size_t row;
  size_t col;

  for(row=0; row<rows; ++row) {
    for(col=0; col<cols; ++col) {
      l_head[index_l_l2(row, col)] = m[index_m(row, col)];
    }
  }
}

void
save_u_panel_l2(double * restrict m,
		const double * restrict u_l2,
		size_t rows,
		size_t cols)
{
  size_t col_block;
  size_t col_ofs;
  size_t row;

  for(col_block=0; col_block<cols; col_block+=BLOCK_SIZE*8) {
    for(row=0; row<rows; ++row) {
      for(col_ofs=0; col_ofs<BLOCK_SIZE; ++col_ofs) 
	m[index_m(row, col_block + col_ofs)] = 
	  u_l2[index_u_l2(row, col_block/8 + col_ofs)]; 
    }
  }
}

		
