/**
   Simple DMA linpack benchmark.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "dma-linpack.h"

#include <stdint.h>
#include <stddef.h>

void
init_problem(double * restrict m,
	     size_t num_equations,
	     uint64_t seed)
{
  int random_seed = seed;
  size_t i, j;

  /* Init the A matrix. */
  for(i=0; i<num_equations; i++) {
    for(j=0; j<num_equations; j++) {
      random_seed = (int)((long)3125*(long)random_seed % 65536L);
      m[index_m(i, j)] = (random_seed - 32768.0)/16384.0;
    }
  }

  /* Init the B vector. */
  for (i = 0; i<num_equations; i++) 
    m[index_m(i, num_equations)] = 0.0;
  
  for (j = 0; j<num_equations; j++)
    for (i = 0; i<num_equations; i++)
      m[index_m(i, num_equations)] += m[index_m(j, i)];
}
