	;; Linpack Block Update
	;;
	;; Calculate M_new = M_old - L*U

;; Copyright (c) 2012 Kungliga Tekniska Högskolan 
;; (Royal Institute of Technology, Stockholm, Sweden).
;; All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions
;; are met:
;;
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;;
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;;
;; THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;; ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
;; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
;; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;; SUCH DAMAGE.

	
	.text
	.global block_update

	;; C - prototype
	;; void block_update(double *mnew, double *mold,
	;;                   double *l, double *u,
	;;                   unsigned block_count,
	;;                   unsigned panel_count)

	.asg "A4", arg_mnew
	.asg "B4", arg_mold
	.asg "A6", arg_l
	.asg "B6", arg_u
	.asg "A8", arg_block_count
	.asg "B8", arg_panel_count


	.asg "A2", mi_a
	.asg "B2", mi_b
	.asg "A3", mo_a
	.asg "B3", mo_b
	.asg "A4", l_a
	.asg "B4", l_b
	.asg "A5", u_a
	.asg "B5", u_b
	.asg "A14", panel_cnt
	.asg "A15", block_cnt
	.asg "B14", panel_reload
	.asg "B15", SP

block_update:
	;; Layout for matrices:
	;; M 16x16 not blocked...
	
	;; Setup stack frame and save stuff
	;; SP MUST ALWAYS be aligned to 16-byte
	;; We must save: A10-A15, B10-B14, B3 (RP)
	
		stdw	A11:A10, *--SP(48) 		; Make frame 48 bytes
		stdw	A13:A12, *+SP(8)
		stdw	A15:A14, *+SP(16)
		stdw	B11:B10, *+SP(24)
		stdw	B13:B12, *+SP(32)
		stw	B14, *+SP(40)
		stw	B3, *+SP(44)

	;; Setup for the first iteration

		mv	arg_block_count, block_cnt
		add	arg_panel_count, -4, panel_cnt
		mv	arg_panel_count, panel_reload

	;; Setup pointers to input matrix (mi_a = mold, mi_b = mold + 1) 
		mv	arg_mold, mi_a
		add	arg_mold, 8, mi_b

	;; Setup pointers to output matrix (mo_a = mnew, mo_b = mnew + 1)
		mv	arg_mnew, mo_a
		add	arg_mnew, 8, mo_b
	
	;; Setup pointers to l panel (l_a = arg_l, l_b = arg_l + 1)
		mv	arg_l, l_a
		add	arg_l, 8, l_b
	
	;; Setup pointers to u panel (u_a = arg_u, u_b = arg_u + 1)
		mv	arg_u, u_a
		add	arg_u, 8, u_b
	
	;; 	add	arg_mnew, 8, mi_b 
	;; ||	mv	arg_mnew, mi_a
	;; ||	mv	arg_mold, mo_a
	;; ||	add	arg_mold, 8, mo_b
	;; ||	mv	arg_l, l_a
	;; ||	mv	arg_panel_count, panel_reload

	;; 	add	arg_l, 8, l_b
	;; ||	mv	arg_u, u_a
	;; ||	add	arg_u, 8, u_b
	;; ||	mv	arg_block_count, block_cnt

	;; 	mv	panel_reload, panel_cnt

	;; Dummy setup for m

		lddw	*mi_a++[2], A17:A16 		; a00=mi
	||	lddw	*mi_b++[2], B17:B16		; a01=mi

		lddw	*mi_a++[14], A19:A18 		; a02=mi
	||	lddw	*mi_b++[14], B19:B18		; a03=mi

		lddw	*mi_a++[2], A21:A20 		; a10=mi
	||	lddw	*mi_b++[2], B21:B20		; a11=mi

		lddw	*mi_a++[14], A23:A22 		; a12=mi
	||	lddw	*mi_b++[14], B23:B22		; a13=mi

		lddw	*mi_a++[2], A25:A24 		; a20=mi
	||	lddw	*mi_b++[2], B25:B24		; a21=mi

		lddw	*mi_a++[14], A27:A26 		; a22=mi
	||	lddw	*mi_b++[14], B27:B26		; a23=mi

		lddw	*mi_a++[2], A29:A28 		; a30=mi
	||	lddw	*mi_b++[2], B29:B28		; a31=mi

		lddw	*mi_a++[14], A31:A30 		; a32=mi
	||	lddw	*mi_b++[14], B31:B30		; a33=mi

		

	;; Iteration 0
		lddw	*l_a++[2], A7:A6		; l0=l[0]
	||	lddw	*l_b++[2], B7:B6		; l1=l[0]

		lddw	*u_a++[2], A9:A8		; u0=u[0]
	||	lddw	*u_b++[2], B9:B8		; u1=u[0]

		lddw	*u_a++[2], A11:A10		; u2=u[0]
	||	lddw	*u_b++[2], B11:B10		; u3=u[0]
	
		lddw	*l_a++[2], A13:A12		; l2=l[0]
	||	lddw	*l_b++[2], B13:B12		; l3=l[0]

		nop

		zero	A1:A0				; Need two 0s
	||	zero	B1:B0
	
	;; Iteration 1

		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	lddw	*l_a++[2], A7:A6		; l0=l[1]
	||	lddw	*l_b++[2], B7:B6		; l1=l[1]

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A1:A0, A31:A30		; a32-=0 (!)
	||	fsubdp  B31:B30, B1:B0, B31:B30		; a33-=0 (!)
	||	lddw	*u_a++[2], A9:A8		; u0=u[1]
	||	lddw	*u_b++[2], B9:B8		; u1=u[1]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[1]
	||	lddw	*u_b++[2], B11:B10		; u3=u[1]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[1]
	||	lddw	*l_b++[2], B13:B12		; l3=l[1]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13

	;; Dot product loop iteration, pipelined
panel_loop:	
		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||	bdec	panel_loop, panel_cnt

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13

	;; End of panel_loop

	;; Epilogue iteration 0

		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||	shl	panel_reload, 5, A0		; ud_a=pr*32
	||	shl	panel_reload, 5, B0		; ud_b=pr*32
	||	add	panel_reload, -8, panel_cnt	; pc=pr-8

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	sub	u_a, A0, u_a			; u_a-=ud_a
	||	sub	u_b, B0, u_b			; u_b-=ud_b
	||	add	block_cnt, -1, A1		; bn=bc-1

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||[!A1] b 	exit_epilogue

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]
	||	mv 	A1, block_cnt			; bc=bn
	
		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13
	||	stdw	A17:A16, *mo_a++[2] 		; mo=a00
	||	stdw	B17:B16, *mo_b++[2]		; mo=a01

	;; Branch taken to exit_epilogue
	
	;; Epilogue iteration 1

		fmpydp	A9:A8, A7:A6, A17:A16		; i00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B17:B16		; i01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||	stdw	A19:A18, *mo_a++[14] 		; mo=a02
	||	stdw	B19:B18, *mo_b++[14]		; mo=a03

		fmpydp	A11:A10, A7:A6, A19:A18 	; i02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B19:B18		; i03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	stdw	A21:A20, *mo_a++[2] 		; mo=a10
	||	stdw	B21:B20, *mo_b++[2]		; mo=a11

		fmpydp	A9:A8, B7:B6, A21:A20		; i10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B21:B20		; i11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]

		fmpydp	A11:A10, B7:B6, A23:A22		; i12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B23:B22		; i13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A25:A24		; i20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B25:B24		; i21=l2*u1
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A27:A26	; i22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B27:B26	; i23=l2*u3
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A29:A28		; i30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B29:B28		; i31=l3*u1
	||	stdw	A23:A22, *mo_a++[14]		; mo=a12
	||	stdw	B23:B22, *mo_b++[14]		; mo=a13

		fmpydp	A11:A10, B13:B12, A1:A0		; i32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B1:B0		; i33=l3*u3
	||	stdw	A25:A24, *mo_a++[2]		; mo=a20
	||	stdw	B25:B24, *mo_b++[2]		; mo=a21

	;; Epilogue iteration 2
	
		fmpydp	A9:A8, A7:A6, A7:A6		; m00=l0*u0 (!)
	||	fmpydp	B9:B8, A7:A6, B7:B6		; m01=l0*u1
	||	stdw	A27:A26, *mo_a++[14]		; mo=a22
	||	stdw	B27:B26, *mo_b++[14]		; mo=a23

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	stdw	A29:A28, *mo_a++[2]		; mo=a30
	||	stdw	B29:B28, *mo_b++[2]		; mo=a31

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fadddp  A17:A16, A7:A6, A17:A16		; i00+=m00
	||	fadddp	B17:B16, B7:B6, B17:B16		; i01+=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fadddp	A19:A18, A17:A16, A19:A18	; i02+=m02
	||	fadddp	B19:B18, B17:B16, B19:B18	; i03+=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fadddp	A21:A20, A19:A18, A21:A20	; i10+=m10
	||	fadddp	B21:B20, B19:B18, B21:B20	; i11+=m11
	||	stdw	A31:A30, *mo_a++[14]		; mo=a32
	||	stdw	B31:B30, *mo_b++[14]		; mo=a33

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fadddp	A23:A22, A21:A20, A23:A22	; i12+=m12
	||	fadddp  B23:B22, B21:B20, B23:B22	; i13+=m13
	||	lddw	*mi_a++[2], A1:A0		; a00N=mi
	||	lddw	*mi_b++[2], B1:B0		; a01N=mi

	;; Epilogue iteration 3
	
		fmpydp	A9:A8, A7:A6, A7:A6		; m00=l0*u0 (!)
	||	fmpydp	B9:B8, A7:A6, B7:B6		; m01=l0*u1 (!)
	||	fadddp	A25:A24, A23:A22, A25:A24	; i20+=m20
	||	fadddp  B25:B24, B23:B22, B25:B24	; i21+=m21
	||	lddw	*mi_a++[14], A17:A16		; a02N=mi
	||	lddw	*mi_b++[14], B17:B16		; a03N=mi

		fmpydp	A11:A10, A7:A6, A7:A6 		; m02=l0*u2 (!)
	||	fmpydp	B11:B10, A7:A6, B7:B6		; m03=l0*u3 (!)
	||	fadddp	A27:A26, A25:A24, A27:A26	; i22+=m22
	||	fadddp	B27:B26, B25:B24, B27:B26	; i23+=m23
	||	lddw	*mi_a++[2], A19:A18		; a10N=mi
	||	lddw	*mi_b++[2], B19:B18		; a11N=mi

		fmpydp	A9:A8, B7:B6, A7:A6		; m10=l1*u0 (!)
	||	fmpydp	B9:B8, B7:B6, B7:B6		; m11=l1*u1 (!)
	||	fadddp	A29:A28, A27:A26, A29:A28	; i30+=m30
	||	fadddp	B29:B28, B27:B26, B29:B28	; i31+=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2 (!)
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3 (!)
	||	fadddp	A1:A0, A29:A28, A31:A30		; i32+=m32 (!)
	||	fadddp	B1:B0, B29:B28, B31:B30		; i33+=m33 (!)
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A21:A20		; m20=l2*u0 (!)
	||	fmpydp  B9:B8, A13:A12, B21:B20		; m21=l2*u1 (!)
	||	fadddp  A17:A16, A7:A6, A9:A8		; i00+=m00 (!)
	||	fadddp	B17:B16, B7:B6, B9:B8		; i01+=m01 (!)
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A21:A20	; m22=l2*u2 (!)
	||	fmpydp	B11:B10, A13:A12, B21:B20	; m23=l2*u3 (!)
	||	fadddp	A19:A18, A7:A6, A11:A10		; i02+=m02 (!)
	||	fadddp	B19:B18, B7:B6, B11:B10		; i03+=m03 (!)
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A21:A20		; m30=l3*u0 (!)
	||	fmpydp	B9:B8, B13:B12, B21:B20		; m31=l3*u1 (!)
	||	fadddp	A21:A20, A7:A6, A17:A16		; i10+=m10 (!)
	||	fadddp	B21:B20, B7:B6, B17:B16		; i11+=m11 (!)
	||	lddw	*mi_a++[14], A21:A20		; a12N=mi
	||	lddw	*mi_b++[14], B21:B20		; a13N=mi

		fmpydp	A11:A10, B13:B12, A23:A22	; m32=l3*u2 (!)
	||	fmpydp	B11:B10, B13:B12, B23:B22	; m33=l3*u3 (!)
	||	fadddp	A23:A22, A21:A20, A27:A26	; i12+=m12 (!)
	||	fadddp  B23:B22, B21:B20, B27:B26	; i13+=m13 (!)
	||	fsubdp	A1:A0, A9:A8, A17:A16		; a00-=i00
	||	fsubdp	B1:B0, B9:B8, B17:B16		; a01-=i01
	||	lddw	*mi_a++[2], A7:A6		; a20N=mi
	||	lddw	*mi_b++[2], B7:B6		; a21N=mi

	;; Epilogue iteration 4
	
		fmpydp	A9:A8, A7:A6, A23:A22		; m00=l0*u0 (!)
	||	fmpydp	B9:B8, A7:A6, B23:B22		; m01=l0*u1 (!)
	||	fadddp	A25:A24, A21:A20, A29:A28	; i20+=m20 (!)
	||	fadddp  B25:B24, B21:B20, B29:B28	; i21+=m21 (!)
	||	fsubdp	A17:A16, A11:A10, A19:A18	; a02-=i02
	||	fsubdp	B17:B16, B11:B10, B19:B18	; a03-=i03
	||	lddw	*mi_a++[14], A7:A6		; a22N=mi
	||	lddw	*mi_b++[14], B7:B6		; a23N=mi

		fmpydp	A11:A10, A7:A6, A23:A22 	; m02=l0*u2 (!)
	||	fmpydp	B11:B10, A7:A6, B23:B22		; m03=l0*u3 (!)
	||	fadddp	A27:A26, A21:A20, A31:A30	; i22+=m22 (!)
	||	fadddp	B27:B26, B21:B20, B31:B30	; i23+=m23 (!)
	||	fsubdp	A19:A18, A17:A16, A21:A20	; a10-=i10
	||	fsubdp	B19:B18, B17:B16, B21:B20	; a11-=i11
	||	lddw	*mi_a++[2], A7:A6		; a30N=mi
	||	lddw	*mi_b++[2], B7:B6		; a31N=mi

		fmpydp	A9:A8, B7:B6, A25:A24		; m10=l1*u0 (!)
	||	fmpydp	B9:B8, B7:B6, B25:B24		; m11=l1*u1 (!)
	||	fadddp	A29:A28, A21:A20, A17:A16	; i30+=m30 (!)
	||	fadddp	B29:B28, B21:B20, B17:B16	; i31+=m31 (!)
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]

		fmpydp	A11:A10, B7:B6, A27:A26		; m12=l1*u2 (!)
	||	fmpydp	B11:B10, B7:B6, B27:B26		; m13=l1*u3 (!)
	||	fadddp	A31:A30, A23:A22, A31:A30	; i32+=m32 (!)
	||	fadddp	B31:B30, B23:B22, B31:B30	; i33+=m33 (!)
	||	fsubdp	A21:A20, A27:A26, A23:A22	; a12-=i12
	||	fsubdp	B21:B20, B27:B26, B23:B22	; a13-=i13
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A29:A28		; m20=l2*u0 (!)
	||	fmpydp  B9:B8, A13:A12, B29:B28		; m21=l2*u1 (!)
	||	fsubdp  A17:A16, A23:A22, A17:A16	; a00-=m00 (!)
	||	fsubdp	B17:B16, B23:B22, B17:B16	; a01-=m01 (!)
	||	fsubdp	A7:A6, A29:A28, A25:A24		; a20-=i20
	||	fsubdp	B7:B6, B29:B28, B25:B24		; a21-=i21
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A23:A22, A19:A18	; a02-=m02 (!)
	||	fsubdp	B19:B18, B23:B22, B19:B18	; a03-=m03 (!)
	||	fsubdp	A7:A6, A31:A30, A27:A26		; a22-=i22
	||	fsubdp	B7:B6, B31:B30, B27:B26		; a23-=i23
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0 
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1 
	||	fsubdp	A21:A20, A25:A24, A21:A20	; a10-=m10 (!)
	||	fsubdp	B21:B20, B25:B24, B21:B20	; a11-=m11 (!)
	||	fsubdp	A7:A6, A17:A16, A29:A28		; a30-=i30
	||	fsubdp	B7:B6, B17:B16, B29:B28		; a31-=i31
	||	lddw	*mi_a++[14], A27:A26		; a32N=mi
	||	lddw	*mi_b++[14], B27:B26		; a33N=mi

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2 
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3 
	||	fsubdp	A23:A22, A27:A26, A23:A22	; a12-=m12 (!)
	||	fsubdp  B23:B22, B27:B26, B23:B22	; a13-=m13 (!)

	;; Epilogue iteration 5
	
		fmpydp	A9:A8, A7:A6, A29:A28		; m00=l0*u0 (!)
	||	fmpydp	B9:B8, A7:A6, B29:B28		; m01=l0*u1 (!)
	||	fsubdp	A25:A24, A29:A28, A25:A24	; a20-=m20 (!)
	||	fsubdp  B25:B24, B29:B28, B25:B24	; a21-=m21 (!)

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22 
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23 

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30 
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31 
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||	b	panel_loop

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A27:A26, A29:A28, A17:A16	; a32-=m32 (!N)
	||	fsubdp	B27:B26, B29:B28, B17:B16	; a33-=m33 (!N)
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A29:A28, A17:A16	; a00-=m00 (!)
	||	fsubdp	B17:B16, B29:B28, B17:B16	; a01-=m01 (!)
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11
	||	fsubdp	A17:A16, A31:A30, A31:A30	; a32-=i32
	||	fsubdp	B17:B16, B31:B30, B31:B30	; a33-=i33

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13

	;; Branch taken to panel_loop

exit_epilogue:
	;; Exit iteration 1
		fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||	stdw	A19:A18, *mo_a++[14] 		; mo=a02
	||	stdw	B19:B18, *mo_b++[14]		; mo=a03

		fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	stdw	A21:A20, *mo_a++[2] 		; mo=a10
	||	stdw	B21:B20, *mo_b++[2]		; mo=a11

		fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	stdw	A23:A22, *mo_a++[14] 		; mo=a12
	||	stdw	B23:B22, *mo_b++[14]		; mo=a13

		fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	stdw	A25:A24, *mo_a++[2] 		; mo=a20
	||	stdw	B25:B24, *mo_b++[2]		; mo=a21

		stdw	A27:A26, *mo_a++[14] 		; mo=a22
	||	stdw	B27:B26, *mo_b++[14]		; mo=a23

		stdw	A29:A28, *mo_a++[2] 		; mo=a30
	||	stdw	B29:B28, *mo_b++[2]		; mo=a31

		stdw	A31:A30, *mo_a++[14] 		; mo=a32
	||	stdw	B31:B30, *mo_b++[14]		; mo=a33
	
	;; Clean-up & restore registers

cleanup:
		ldw	*+SP(44), B3
		ldw	*+SP(40), B14
		lddw	*+SP(32), B13:B12
		lddw	*+SP(24), B11:B10
		lddw	*+SP(16), A15:A14
		lddw	*+SP(8), A13:A12
		lddw	*SP++(48), A11:A10
		bnop	B3, 5
	;; Return from function here
	