	;; Linpack Block Update
	;;
	;; Calculate M_new = M_old - L*U

;; Copyright (c) 2012 Kungliga Tekniska Högskolan 
;; (Royal Institute of Technology, Stockholm, Sweden).
;; All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions
;; are met:
;;
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;;
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;;
;; THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;; ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
;; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
;; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;; SUCH DAMAGE.

	.text

	.global update_block_idma_new

	;; C - prototype
	;; unsigned update_block_idma_new(double * restrict m_l2,
	;;       		          const double * restrict u_l2,
	;; 			          const double * restrict l_l2,
	;; 			          double * restrict m_l1,
	;; 			          double * restrict u_l1,
	;; 			          double * restrict l_l1,
	;; 			          unsigned counter_reload,
	;; 			          unsigned block_cols,
	;; 			          unsigned panel_size);
	;;
	;; BLOCK_SIZE is fixed to 16x16 elements
	;; M matrix MUST be multiples of block size
	;; block_rows and block_cols give sizes in BLOCKS
	;; panel_size is size in elements, MUST BE >=16 and <=255
	;;
	;; On Entry: the matrices m_l1, u_l1 and l_l1 MUST BE initialised with
	;;           the first block and panels that we want to calculate on
	;;
	;;           the m_l2, u_l2 and l_l2 pointers should point to the start
	;;           of the blocks in the L2 cache instead.
	;;
	;; m_l2_load: Increments with 1/4 BS each transfer
	;;            always 1 block ahead of mo
	;; m_l1_load: swap with m_l1_calc after one block
	;;
	;; u_l1_load: swap with u_l1_calc after each sub column
	;;
	;; u_l2_load: increment with each transfer
	;; 	      reset at end of each row
	;;
	;; l_l2_load: increment with each transfer
	;; 	      reset to begin at half of block update, except end of
	;; 	      current row
	;; l_l1_load: increment with each transfer
	;; 	      swap at end of half block
	;;
	;; mi_a, mi_b:
	;;
	;; l_a, l_b:  panel_loop l_a += 4 (total l_a += panel_size*4)
	;; 	      block_cnt odd no adjust
	;;            block_cnt 14,12,10, 8 l_a -= 8*panel_size
	;;            block_cnt  6, 4, 2, 0 no adjust
	;;
	;; u_a, u_b:  panel_loop u_a += 4 (total u_a += panel_size*4)
	;;            block_cnt odd u_a -= 4*panel_size (reset to start)
	;;            block_cnt

;;; General comments on data layout & loop-order
;;;
;;; within a 16x16 block (4x4 sub-blocks) the order is following
;;; block_cnt
;;; 15 13 11  9
;;; 14 12 10  8
;;;  7  5  3  1
;;;  6  4  2  0
;;; 
;;; new         Offset	Delta	Offset	Delta	Offset	Delta
;;; block_cnt	m_l1		l_l1		u_l1
;;; 15		0,0		0	+1	0	 
;;; 14		1,0		1	-1	0	+1
;;; 13		0,1		0	+1	1
;;; 12		1,1		1	-1	1	-1
;;; 11		0,2		0	+1	0
;;; 10		1,2		1	-1	0	+1
;;;  9		0,3		0	+1	1
;;;  8		1,3		1	+1	1	-1
;;;  7		2,0		2	+1	0
;;;  6		3,0		3	-1	0	+1
;;;  5		2,1		2	+1	1
;;;  4		3,1		3	-1	1	-1
;;;  3		2,2		2	+1	0
;;;  2		3,2		3	-1	0	+1
;;;  1		2,3		2	+1	1
;;;  0		3,3		3	-3	1	-1
;;;
;;;
;;; Update Rules ideas:
;;; 
;;; u_l1 : nbc even         -> u_l1 -= panel_size*4*8
;;;        nbc 15, 11, 7, 3 -> u_l1 -= panel_size*8*8
;;; 	   nbc 13,  9, 5, 1 -> no adjustment (next col)
;;;
;;; l_l1 : nbc even               -> no adjustment (next row)
;;;        nbc 15                 -> l_l1 -= panel_size*16*8
;;;        nbc  7                 -> no adjustment (next row)
;;;        nbc 13, 11, 9, 5, 3, 1 -> l_l1 -= panel_size*8*8
;;;
;;; General idea on loop counter:
;;; split 32-bit register into 4 8-bit fields
;;; row | col | blk | pnl
;;;
;;; Counter decrements, on underflow of fields, next right field gets updated
;;; Check counter before decrement -> if !0 continue in loop
;;; On underflow -> we have all ones in the least significant fields (-1)
;;; For panel loop:
;;;       add     loop_cnt, -1, loop_cnt
;;;       cmpgtu4 loop_cnt, loop_rld, A0
;;; [!A0] b       panel_loop
;;;       dminu4  loop_cnt, loop_rld, loop_cnt 
;;; 
	
	;; Arguments
	.asg "A4",  arg_m_l2
	.asg "B4",  arg_u_l2
	.asg "A6",  arg_l_l2
	.asg "B6",  arg_m_l1
	.asg "A8",  arg_u_l1
	.asg "B8",  arg_l_l1
	.asg "A10", arg_cnt_reload
	
	;; Register assignments
	.asg "A2", mi_a
	.asg "B2", mi_b
	.asg "A3", mo_a
	.asg "B3", mo_b
	.asg "A4", l_a
	.asg "B4", l_b
	.asg "A5", u_a
	.asg "B5", u_b
	.asg "A14", panel_cnt
	.asg "B14", idma1_base
	.asg "A15", panel_reload
	.asg "B15", SP
	
	;; Local Variables on the Stack (ofs 0 - 63)
 	.asg "*+SP( 0)", u_l2_start
 	.asg "*+SP( 4)", u_l2_load
 	.asg "*+SP( 8)", m_l2_load
 	.asg "*+SP(12)", l_l2_load
 	;; .asg "*+SP(16)", l_l2_load
 	;; .asg "*+SP(20)", m_l1_load
 	;; .asg "*+SP(24)", m_l2_load
 	;; .asg "*+SP(28)", l_l1_calc
 	;; .asg "*+SP(32)", l_l2_next
	;; .asg "*+SP(36)", block_rows
	;; .asg "*+SP(40)", block_cols
	;; .asg "*+SP(44)", block_row_cnt
	;; .asg "*+SP(48)", block_col_cnt

 	.asg "0x01820100", idma1_base_addr
 	.asg "0x00000200", m_load_cnt
 	.asg "0xFFFFC000", u_block_adjust
	.asg "0x00000800", m_block_size


update_block_idma_new:
	.asmfunc
 	;; Setup stack frame and save stuff
 	;; SP MUST ALWAYS be aligned to 16-byte
 	;; We must save: A10-A15, B10-B14, B3 (RP)

 		subaw	SP, 28, SP 			; Make frame 112 bytes
	
 		stdw	A11:A10, *+SP(64) 		; Registers saved 64+
 		stdw	A13:A12, *+SP(72)
 		stdw	A15:A14, *+SP(80)
 		stdw	B11:B10, *+SP(88)
 		stdw	B13:B12, *+SP(96)
 		stw	B14, *+SP(104)
 		stw	B3, *+SP(108)

	;; Init local variables

		extu	arg_cnt_reload, 24, 19, A13 	; Extract ps*32
		add	arg_u_l2, A13, B5		
		stdw	B5:B4, u_l2_start 		; Init current u_l2 ptr

		mvk	m_block_size - m_load_cnt, A0
		add	arg_m_l2, A0, A0  		; m_l2_load+=bs*bs*8
		add	arg_l_l2, A13, A1 		; l_l2_load+=ps*32
		stdw	A1:A0, m_l2_load		; Init m_l2/l_l2 ptr

	;; Init register variables
		mv	arg_m_l1, mi_a
	||	add	arg_m_l1, 8, mi_b

		mv	arg_m_l2, mo_a
	||	add	arg_m_l2, 8, mo_b

		mv	arg_u_l1, u_a
	||	add	arg_u_l1, 8, u_b

		mv	arg_l_l1, l_a
	||	add	arg_l_l1, 8, l_b

		mv	arg_cnt_reload, panel_reload
	||	mv	arg_cnt_reload, panel_cnt
		add	panel_cnt, -4, panel_cnt

		mvkl	idma1_base_addr, idma1_base
		mvkh	idma1_base_addr, idma1_base

panel_restart:	
	;; Dummy setup for m

		lddw	*mi_a++[2], A17:A16 		; a00=mi
	||	lddw	*mi_b++[2], B17:B16		; a01=mi

		lddw	*mi_a++[14], A19:A18 		; a02=mi
	||	lddw	*mi_b++[14], B19:B18		; a03=mi

		lddw	*mi_a++[2], A21:A20 		; a10=mi
	||	lddw	*mi_b++[2], B21:B20		; a11=mi

		lddw	*mi_a++[14], A23:A22 		; a12=mi
	||	lddw	*mi_b++[14], B23:B22		; a13=mi

		lddw	*mi_a++[2], A25:A24 		; a20=mi
	||	lddw	*mi_b++[2], B25:B24		; a21=mi

		lddw	*mi_a++[14], A27:A26 		; a22=mi
	||	lddw	*mi_b++[14], B27:B26		; a23=mi

		lddw	*mi_a++[2], A29:A28 		; a30=mi
	||	lddw	*mi_b++[2], B29:B28		; a31=mi

		lddw	*mi_a++[14], A31:A30 		; a32=mi
	||	lddw	*mi_b++[14], B31:B30		; a33=mi

	;; Iteration 0
		lddw	*l_a++[2], A7:A6		; l0=l[0]
	||	lddw	*l_b++[2], B7:B6		; l1=l[0]

		lddw	*u_a++[2], A9:A8		; u0=u[0]
	||	lddw	*u_b++[2], B9:B8		; u1=u[0]

		lddw	*u_a++[2], A11:A10		; u2=u[0]
	||	lddw	*u_b++[2], B11:B10		; u3=u[0]
	
		lddw	*l_a++[2], A13:A12		; l2=l[0]
	||	lddw	*l_b++[2], B13:B12		; l3=l[0]

		nop

		zero	A1:A0				; Need two 0s
	||	zero	B1:B0
	
	;; Iteration 1

		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	lddw	*l_a++[2], A7:A6		; l0=l[1]
	||	lddw	*l_b++[2], B7:B6		; l1=l[1]

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A1:A0, A31:A30		; a32-=0 (!)
	||	fsubdp  B31:B30, B1:B0, B31:B30		; a33-=0 (!)
	||	lddw	*u_a++[2], A9:A8		; u0=u[1]
	||	lddw	*u_b++[2], B9:B8		; u1=u[1]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[1]
	||	lddw	*u_b++[2], B11:B10		; u3=u[1]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[1]
	||	lddw	*l_b++[2], B13:B12		; l3=l[1]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13

	;; Dot product loop iteration, pipelined
panel_loop:	
		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	shl	panel_cnt, 24, A0		; Extract panel part

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||[A0] 	b	panel_loop			; cnt!=0 -> panel_loop
 	||[!A0]	add	panel_cnt, -1, B0		; also copy into B0

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]
	||	add	panel_cnt, -1, panel_cnt	; panel_cnt--

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]


		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13

	;; End of panel_loop

epilogue_start:	
	;; Epilogue iteration N-1
	;; Iteration N-1 of panel_loop

		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||	extu.S	panel_reload, 24, 18, A1	; extract ps*64
	||	extu.S	B0, 23, 31, B0			; bi_new even?

	;; L1 Unit MUST NOT be used for DP op here (minu4)!
		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp.S A27:A26, A25:A24, A27:A26	; a22-=m22 (UNIT!)
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	add	panel_cnt, 1, A0		; panel_cnt + 1 (old)
 	||[B0]	ldw	*idma1_base, B0			; Check IDMA1 status
;;	||	zero	B0

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||[!A0] b.S	exit_epilogue
	||	extu.S	A0, 20, 28, A0 			; extract bi_old

	;; S1 Writeback here! Need WB free L unit!
		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp.S A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]
	||	minu4.L panel_cnt, panel_reload, panel_cnt ; Adjust counters
	||	mv	A1, B1				; l_adj=2*ps

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]
	||	sub	A0, 8, A0			; bi==8
	||[!A0] add	B1, B1, B1			; l_adj=4*ps (bi==0)

	;; S1 Writeback here!
		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]
	||	add.S	A0, 8, A0			; restore bi
	||[!A0]	zero.S	B1				; l_adj=0 (bi==8)

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11
	||	and	A0, 3, A0			; ud=bi[1..0]
 
		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13
	||	and	A0, 1, A1			; lu=bi[0]
	||[!A0]	mv	A1, B12				; u_adj=2*ps (ud==0)
	||[A0]	shr	A1, 1, B12			; u_adj=ps (ud!=0)
	
	;; Branch taken to exit_loop
	
 	;; Epilogue iteration 0

		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||	sub	A0, 2, A0			; ud==2
	||[!A1]	sub	l_a, B1, l_a			; l_a-=l_adj (bi[0]==0)
	||[!A1]	sub	l_b, B1, l_b			; l_b-=l_adj (bi[0]==0)

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||[A0]	sub	u_a, B12, u_a			; u_a-=u_adj (ud!=2)
	||[A0]	sub	u_b, B12, u_b			; u_b-=u_adj (ud!=2)

	  [!B0]	fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||[!B0]	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||[!B0]	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||[!B0]	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||	add	panel_cnt, -9, panel_cnt	; Adjust panel count
	||[B0]	b	idma_wait

	  [!B0]	fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||[!B0]	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||[!B0]	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||[!B0]	lddw	*u_b++[2], B9:B8		; u1=u[i]
	
	  [!B0]	fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||[!B0]	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||[!B0]	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||[!B0]	lddw	*u_b++[2], B11:B10		; u3=u[i]

	  [!B0]	fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||[!B0]	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||[!B0]	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||[!B0]	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||[!B0]	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||[!B0]	lddw	*l_b++[2], B13:B12		; l3=l[i]

	  [!B0]	fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||[!B0]	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||[!B0]	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||[!B0]	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

	  [!B0]	fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||[!B0]	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||[!B0]	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||[!B0]	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13
	||	stdw	A17:A16, *mo_a++[2] 		; mo=a00
	||	stdw	B17:B16, *mo_b++[2]		; mo=a01

	;; Branch taken to idma_wait
		
	;; Epilogue iteration 1
epilogue_resume:	
		fmpydp	A9:A8, A7:A6, A17:A16		; i00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B17:B16		; i01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||	stdw	A19:A18, *mo_a++[14] 		; mo=a02
	||	stdw	B19:B18, *mo_b++[14]		; mo=a03
	||	sub	panel_reload, panel_cnt, A18	; ofs = p_r - p_c

		fmpydp	A11:A10, A7:A6, A19:A18 	; i02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B19:B18		; i03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	stdw	A21:A20, *mo_a++[2] 		; mo=a10
	||	stdw	B21:B20, *mo_b++[2]		; mo=a11
	||	shr	A18, 8, A19			; row = ofs >> 8
	||	mvk	0xfffff000, B1			; cyc_adj=-4096

		fmpydp	A9:A8, B7:B6, A21:A20		; i10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B21:B20		; i11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||	and	A19, 15, B0			; bi==15 (b_ofs==0)
	||	andn	A19, panel_reload, A19		; row_c=row & ~p_r(c)

		fmpydp	A11:A10, B7:B6, A23:A22		; i12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B23:B22		; i13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]
	||	xor	A18, A19, A18			; cor = row_c ^ ofs
	||[B0]	zero	B1				; cyc_adj=0 (bi!=15)

		fmpydp	A9:A8, A13:A12, A25:A24		; i20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B25:B24		; i21=l2*u1
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]
	||	extu	A18, 15, 31, A0			; extract cor[16] only

		fmpydp	A11:A10, A13:A12, A27:A26	; i22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B27:B26	; i23=l2*u3
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]
	||[!A0]	add	mi_a, B1, mi_a			; mi_a+=cyc_adj
	||[!A0]	add	mi_b, B1, mi_b			; mi_b+=cyc_adj

		fmpydp	A9:A8, B13:B12, A29:A28		; i30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B29:B28		; i31=l3*u1
	||	stdw	A23:A22, *mo_a++[14]		; mo=a12
	||	stdw	B23:B22, *mo_b++[14]		; mo=a13
	||	extu	panel_cnt, 21, 29, A0		; extract bi[2..0]

		fmpydp	A11:A10, B13:B12, A1:A0		; i32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B1:B0		; i33=l3*u3
	||	stdw	A25:A24, *mo_a++[2]		; mo=a20
	||	stdw	B25:B24, *mo_b++[2]		; mo=a21
	||	sub	A0, 7, A0			; bi==7 || bi==15
	||	mvk	0xfffffc20, A1			; mo_adj = -8*r + 4*c

	;; Epilogue iteration 2
	
		fmpydp	A9:A8, A7:A6, A7:A6		; m00=l0*u0 (!)
	||	fmpydp	B9:B8, A7:A6, B7:B6		; m01=l0*u1
	||	stdw	A27:A26, *mo_a++[14]		; mo=a22
	||	stdw	B27:B26, *mo_b++[14]		; mo=a23
	||[!A0]	mvk	0xffffffa0, A1			; mo_adj=-12*c (bi==7)
	||	and	A0, 1, A0			; bi even (bi-7 odd)?

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	stdw	A29:A28, *mo_a++[2]		; mo=a30
	||	stdw	B29:B28, *mo_b++[2]		; mo=a31
	||[A0]	zero	A1				; m_adj=0 (bi even)
	
		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||	dmvd.S	A0, A1, A13:A12			; move m_adj
	||	add	mi_a, A1, mi_a			; mi_a+=m_adj
	||	add	mi_b, A1, mi_b			; mi_b+=m_adj

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fadddp  A17:A16, A7:A6, A17:A16		; i00+=m00
	||	fadddp	B17:B16, B7:B6, B17:B16		; i01+=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fadddp	A19:A18, A17:A16, A19:A18	; i02+=m02
	||	fadddp	B19:B18, B17:B16, B19:B18	; i03+=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fadddp	A21:A20, A19:A18, A21:A20	; i10+=m10
	||	fadddp	B21:B20, B19:B18, B21:B20	; i11+=m11
	||	stdw	A31:A30, *mo_a++[14]		; mo=a32
	||	stdw	B31:B30, *mo_b++[14]		; mo=a33

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fadddp	A23:A22, A21:A20, A23:A22	; i12+=m12
	||	fadddp  B23:B22, B21:B20, B23:B22	; i13+=m13
	||	lddw	*mi_a++[2], A1:A0		; a00N=mi
	||	lddw	*mi_b++[2], B1:B0		; a01N=mi
	||	add.S	mo_a, A12, mo_a			; mo_a+=m_adj
	||	add.S	mo_b, A12, mo_b			; mo_b+=m_adj

	;; Epilogue iteration 3
	
		fmpydp	A9:A8, A7:A6, A7:A6		; m00=l0*u0 (!)
	||	fmpydp	B9:B8, A7:A6, B7:B6		; m01=l0*u1 (!)
	||	fadddp	A25:A24, A23:A22, A25:A24	; i20+=m20
	||	fadddp  B25:B24, B23:B22, B25:B24	; i21+=m21
	||	lddw	*mi_a++[14], A17:A16		; a02N=mi
	||	lddw	*mi_b++[14], B17:B16		; a03N=mi

		fmpydp	A11:A10, A7:A6, A7:A6 		; m02=l0*u2 (!)
	||	fmpydp	B11:B10, A7:A6, B7:B6		; m03=l0*u3 (!)
	||	fadddp	A27:A26, A25:A24, A27:A26	; i22+=m22
	||	fadddp	B27:B26, B25:B24, B27:B26	; i23+=m23
	||	lddw	*mi_a++[2], A19:A18		; a10N=mi
	||	lddw	*mi_b++[2], B19:B18		; a11N=mi

		fmpydp	A9:A8, B7:B6, A7:A6		; m10=l1*u0 (!)
	||	fmpydp	B9:B8, B7:B6, B7:B6		; m11=l1*u1 (!)
	||	fadddp	A29:A28, A27:A26, A29:A28	; i30+=m30
	||	fadddp	B29:B28, B27:B26, B29:B28	; i31+=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2 (!)
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3 (!)
	||	fadddp	A1:A0, A29:A28, A31:A30		; i32+=m32 (!)
	||	fadddp	B1:B0, B29:B28, B31:B30		; i33+=m33 (!)
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A21:A20		; m20=l2*u0 (!)
	||	fmpydp  B9:B8, A13:A12, B21:B20		; m21=l2*u1 (!)
	||	fadddp  A17:A16, A7:A6, A9:A8		; i00+=m00 (!)
	||	fadddp	B17:B16, B7:B6, B9:B8		; i01+=m01 (!)
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A21:A20	; m22=l2*u2 (!)
	||	fmpydp	B11:B10, A13:A12, B21:B20	; m23=l2*u3 (!)
	||	fadddp	A19:A18, A7:A6, A11:A10		; i02+=m02 (!)
	||	fadddp	B19:B18, B7:B6, B11:B10		; i03+=m03 (!)
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A21:A20		; m30=l3*u0 (!)
	||	fmpydp	B9:B8, B13:B12, B21:B20		; m31=l3*u1 (!)
	||	fadddp	A21:A20, A7:A6, A17:A16		; i10+=m10 (!)
	||	fadddp	B21:B20, B7:B6, B17:B16		; i11+=m11 (!)
	||	lddw	*mi_a++[14], A21:A20		; a12N=mi
	||	lddw	*mi_b++[14], B21:B20		; a13N=mi

		fmpydp	A11:A10, B13:B12, A23:A22	; m32=l3*u2 (!)
	||	fmpydp	B11:B10, B13:B12, B23:B22	; m33=l3*u3 (!)
	||	fadddp	A23:A22, A21:A20, A27:A26	; i12+=m12 (!)
	||	fadddp  B23:B22, B21:B20, B27:B26	; i13+=m13 (!)
	||	fsubdp	A1:A0, A9:A8, A17:A16		; a00-=i00
	||	fsubdp	B1:B0, B9:B8, B17:B16		; a01-=i01
	||	lddw	*mi_a++[2], A7:A6		; a20N=mi
	||	lddw	*mi_b++[2], B7:B6		; a21N=mi

	;; Epilogue iteration 4
	
		fmpydp	A9:A8, A7:A6, A23:A22		; m00=l0*u0 (!)
	||	fmpydp	B9:B8, A7:A6, B23:B22		; m01=l0*u1 (!)
	||	fadddp	A25:A24, A21:A20, A29:A28	; i20+=m20 (!)
	||	fadddp  B25:B24, B21:B20, B29:B28	; i21+=m21 (!)
	||	fsubdp	A17:A16, A11:A10, A19:A18	; a02-=i02
	||	fsubdp	B17:B16, B11:B10, B19:B18	; a03-=i03
	||	lddw	*mi_a++[14], A7:A6		; a22N=mi
	||	lddw	*mi_b++[14], B7:B6		; a23N=mi

		fmpydp	A11:A10, A7:A6, A23:A22 	; m02=l0*u2 (!)
	||	fmpydp	B11:B10, A7:A6, B23:B22		; m03=l0*u3 (!)
	||	fadddp	A27:A26, A21:A20, A31:A30	; i22+=m22 (!)
	||	fadddp	B27:B26, B21:B20, B31:B30	; i23+=m23 (!)
	||	fsubdp	A19:A18, A17:A16, A21:A20	; a10-=i10
	||	fsubdp	B19:B18, B17:B16, B21:B20	; a11-=i11
	||	lddw	*mi_a++[2], A7:A6		; a30N=mi
	||	lddw	*mi_b++[2], B7:B6		; a31N=mi

		fmpydp	A9:A8, B7:B6, A25:A24		; m10=l1*u0 (!)
	||	fmpydp	B9:B8, B7:B6, B25:B24		; m11=l1*u1 (!)
	||	fadddp	A29:A28, A21:A20, A17:A16	; i30+=m30 (!)
	||	fadddp	B29:B28, B21:B20, B17:B16	; i31+=m31 (!)
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]

		fmpydp	A11:A10, B7:B6, A27:A26		; m12=l1*u2 (!)
	||	fmpydp	B11:B10, B7:B6, B27:B26		; m13=l1*u3 (!)
	||	fadddp	A31:A30, A23:A22, A31:A30	; i32+=m32 (!)
	||	fadddp	B31:B30, B23:B22, B31:B30	; i33+=m33 (!)
	||	fsubdp	A21:A20, A27:A26, A23:A22	; a12-=i12
	||	fsubdp	B21:B20, B27:B26, B23:B22	; a13-=i13
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A29:A28		; m20=l2*u0 (!)
	||	fmpydp  B9:B8, A13:A12, B29:B28		; m21=l2*u1 (!)
	||	fsubdp  A17:A16, A23:A22, A17:A16	; a00-=m00 (!)
	||	fsubdp	B17:B16, B23:B22, B17:B16	; a01-=m01 (!)
	||	fsubdp	A7:A6, A29:A28, A25:A24		; a20-=i20
	||	fsubdp	B7:B6, B29:B28, B25:B24		; a21-=i21
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A23:A22, A19:A18	; a02-=m02 (!)
	||	fsubdp	B19:B18, B23:B22, B19:B18	; a03-=m03 (!)
	||	fsubdp	A7:A6, A31:A30, A27:A26		; a22-=i22
	||	fsubdp	B7:B6, B31:B30, B27:B26		; a23-=i23
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0 
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1 
	||	fsubdp	A21:A20, A25:A24, A21:A20	; a10-=m10 (!)
	||	fsubdp	B21:B20, B25:B24, B21:B20	; a11-=m11 (!)
	||	fsubdp	A7:A6, A17:A16, A29:A28		; a30-=i30
	||	fsubdp	B7:B6, B17:B16, B29:B28		; a31-=i31
	||	lddw	*mi_a++[14], A27:A26		; a32N=mi
	||	lddw	*mi_b++[14], B27:B26		; a33N=mi

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2 
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3 
	||	fsubdp	A23:A22, A27:A26, A23:A22	; a12-=m12 (!)
	||	fsubdp  B23:B22, B27:B26, B23:B22	; a13-=m13 (!)

	;; Epilogue iteration 5

	;; Writeback on S units here!
		fmpydp	A9:A8, A7:A6, A29:A28		; m00=l0*u0 (!)
	||	fmpydp	B9:B8, A7:A6, B29:B28		; m01=l0*u1 (!)
	||	fsubdp	A25:A24, A29:A28, A25:A24	; a20-=m20 (!)
	||	fsubdp  B25:B24, B29:B28, B25:B24	; a21-=m21 (!)

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22 
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	extu.S	panel_cnt, 23, 31, A0		; bi even?

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30 
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31 
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||[A0]	b	epilogue_idma			; need to setup IDMA1
	||[A0]	add	panel_cnt, -4, panel_cnt	; Adjust panel counter

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A27:A26, A29:A28, A17:A16	; a32-=m32 (!N)
	||	fsubdp	B27:B26, B29:B28, B17:B16	; a33-=m33 (!N)
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A29:A28, A17:A16	; a00-=m00 (!)
	||	fsubdp	B17:B16, B29:B28, B17:B16	; a01-=m01 (!)
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]
	
		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11
	||	fsubdp	A17:A16, A31:A30, A31:A30	; a32-=i32
	||	fsubdp	B17:B16, B31:B30, B31:B30	; a33-=i33

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13

	;; Branch to epilogue_idma taken here

	;; Epilogue iteration 6

	;; ! Writeback on S units 
		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	|| 	b	panel_loop			; Back to normal loop

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13
	
	;; Branch taken to panel_loop

epilogue_idma:	
		;; Epilogue iteration 6 (IDMA)
	
	;; ! Writeback on S units 
		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||	lddw	u_l2_start, B7:B6		; Get L2 pointers
	||	subaw	u_a, 24, A0
	
		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	extu	panel_reload, 24, 19, A0	; panel_sz=p_r*32
	||	subaw	A0, 24, A6

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||	extu	panel_cnt, 16, 24, A1		; Extract bi
	||	mv	A0, B0				; l2_adj=ps

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]
	||	shl	A0, 2, B1			; 4*ps
	||	sub	A1, 9, A1			; bi==9

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]
	||	extu	panel_cnt, 8, 16, A1		; Extract bi and ci
	||[!A1]	sub	B0, B1, B0			; l2_adj-=4*ps

	;; U L2 pointers now in B7:B6
		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]
	||	add	B0, B7, B0			; Adjust src ptr
	||	sub	A1, 1, A1			; bi==1 & ci==0

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11
	||	extu	panel_cnt, 22, 31, A1		; Extract bi[1]
	||[!A1]	mv	B6, B0				; Reset to start
	||	mv	A6, A12
	||	lddw	m_l2_load, A7:A6		; Get m_l2 & l_l2 ptrs

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13
	||	stw	B0, *+idma1_base(8)		; Write out source
	||[!A1]	neg	A0, A0				; adj=-ps
	
	;; Epilogue iteration 7 (IDMA)
	
		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||	stw	B0, u_l2_load			; Save L2 ptr
	||	add	A0, A12, A1			; dst=L1+adj
	||[!A1]	neg	A0, A0				; sz=ps

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	stw	A1, *+idma1_base(12)		; Write out dest
	||	extu	panel_cnt, 8, 16, A1		; Extract bi and ci

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||	sub	A1, 3, A1			; ci==0 & bi==3
	||	shr	B1, 2, B0			; adj=ps

	;; m_l2_load in A6, l_l2_load in A7
		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]
	||[!A1]	add	B0, B1, B0			; adj+=4*ps (ci==0 &...)
	||	and	A1, 15, A1			; Extract bi-3

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]
	||[!A1]	sub	B0, B1, B0			; adj-=4*ps (bi==3)
	||	add	A1, 3, A1			; Regen bi

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]
	||	mvkh	0xe0000000, A0			; Set IDMA priority 7
	||	shr	B1, 2, B1			; B1: ps

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11
 	||	stw	A0, *+idma1_base(16)		; Write out size (U)
	||	add	B0, A7, B0			; l_l2_load+=adj
	||	and	A1, 4, A1			; (bi & 4)==0
	||	mv	A6, A12				; move away m_l2_load

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13
	||[!A1]	stw	B0, l_l2_load			; Update l_l2 ptr
	||[A1]	zero	A0				; Set count=0 if no xfer
	||	addk	m_load_cnt, A12			; Adjust m_l2 ptr
	
	;; Epilogue iteration 8 (IDMA)
	
		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||[!A0]	stw	A12, m_l2_load			; Store m_l2 ptr
	||	mvk	1024, A13			; load ofs
	
		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||[!A0]	stw	A12, *+idma1_base(8)		; Write out src
	||	extu	panel_cnt, 22, 31, A1		; Extract bi[1]
	||	add	mi_a, A13, A7			; m_l1=mi_a + 8*rows

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||[!A1]	addk	480, A7				; m_l1+= 4r - 4c

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]
	||	sub	panel_reload, panel_cnt, A1	; ofs=p_r - p_c

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]
 	||	shr	A1, 8, A6			; row=ofs>>8

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]
	||	andn	A6, panel_reload, A13		; row_c=row & ~p_r(c)

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11
	||	mv	A7, A12				; move away m_l1 ptr
	||[A0]	stw	B0, *+idma1_base(8)		; Write out source
 	||	and	A6, 15, B0			; bi==15 (?)
	||	xor	A1, A13, A1			; cor=row^ofs

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13
	||[B0]	mvk	4096, B0			; only unwrap i nec.
	||	extu	A1, 15, 31, A1 			; Extract cor[16]
	
	;; Epilogue iteration 9 (IDMA)
	
		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||[A1]  sub	A12, B0, A12			; unwrap if bi==15

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||[!A0]	stw	A12, *+idma1_base(12)		; Write out dest

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]
	||	mvk	m_load_cnt, A6			; load counter

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]
	||	mvkh	0xe0000000, A6			; Set IDMA priority 7

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11
	||[!A0]	stw	A6, *+idma1_base(16)		; Write out size

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13

	;; Epilogue iteration 10 (IDMA)
	
		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||	extu	panel_cnt, 16, 24, A1		; Extract bi
	||	shl	B1, 1, B0			; adj=2*ps

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	mv	A1, A6				; Save bi
	||	and	A1, 8, A1			; Extract bi[3]

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	|| 	b	panel_loop			; Back to normal loop
	||[!A1]	neg	B0, B0				; adj=-2*ps (bi<8)

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]
	||	and	A6, 2, A1			; Extract bi[1]
	||	addk	-344, B0			; adj-= 11*32 + 8 - 16

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]
	||[!A1]	add	B0, B1, B0			; adj+=ps (bi[1]==0)

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]
	||	add	B0, l_b, B0			; Calculate l1 ptr

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11
	||[A0]	stw	B0, *+idma1_base(12)		; Write out dest

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13
	||[A0]	stw	A0, *+idma1_base(16)		; Write out size
	
	;; Branch taken to panel_loop


idma_wait:
		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
idma_loop:	
	;; We get here from iteration 0 if IDMA1 was still busy...
		ldw	*idma1_base, B0 		; recheck status
		nop	9				; give a little releave
	[B0]	bnop	idma_loop, 5			; still busy?
	
	  	fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||	b	epilogue_resume
	
	  	fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]
	
		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

	  	fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

	  	fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

	  	fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13

	;; Branch taken to epilogue_resume

exit_epilogue:
	;; Exit iteration 0
	
		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13
	||	stdw	A17:A16, *mo_a++[2] 		; mo=a00
	||	stdw	B17:B16, *mo_b++[2]		; mo=a01
	
	;; Exit iteration 1
		fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||	stdw	A19:A18, *mo_a++[14] 		; mo=a02
	||	stdw	B19:B18, *mo_b++[14]		; mo=a03

		fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	stdw	A21:A20, *mo_a++[2] 		; mo=a10
	||	stdw	B21:B20, *mo_b++[2]		; mo=a11

		fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	stdw	A23:A22, *mo_a++[14] 		; mo=a12
	||	stdw	B23:B22, *mo_b++[14]		; mo=a13

		fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	stdw	A25:A24, *mo_a++[2] 		; mo=a20
	||	stdw	B25:B24, *mo_b++[2]		; mo=a21

		stdw	A27:A26, *mo_a++[14] 		; mo=a22
	||	stdw	B27:B26, *mo_b++[14]		; mo=a23

		stdw	A29:A28, *mo_a++[2] 		; mo=a30
	||	stdw	B29:B28, *mo_b++[2]		; mo=a31

		stdw	A31:A30, *mo_a++[14] 		; mo=a32
	||	stdw	B31:B30, *mo_b++[14]		; mo=a33
	
	;; Clean-up & restore registers

cleanup:
		ldw	*+SP(108), B3
		ldw	*+SP(104), B14
		lddw	*+SP(96), B13:B12
		lddw	*+SP(88), B11:B10
		lddw	*+SP(80), A15:A14
		lddw	*+SP(72), A13:A12
		lddw	*+SP(64), A11:A10
		addaw	SP, 28, SP 			; Frame was 112 bytes
		bnop	B3, 5
	;; Return from function here
	.endasmfunc
