	;; Linpack Block Update
	;;
	;; Calculate M_new = M_old - L*U

;; Copyright (c) 2012 Kungliga Tekniska Högskolan 
;; (Royal Institute of Technology, Stockholm, Sweden).
;; All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions
;; are met:
;;
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;;
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;;
;; THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
;; ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
;; ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
;; OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
;; LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
;; OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
;; SUCH DAMAGE.
	
	.text
	.global update_block_l2

	;; C - prototype
	;; void update_block_l2(double *mnew, double *mold,
	;;                   	double *l, double *u,
	;;                   	unsigned block_count,
	;;                   	unsigned panel_count)

	;; void update_block_l2(double *mnew,
	;;                      double *m_l1_calc,
	;;                      double *m_l1_load,
	;;                      double *l_l1,
	;;                      double *l_l2_next,
	;;                      double *u_l1,
	;;                      double *m_l2_load,
	;;                      double *l_l2_load,
	;;                      double *u_l2,
	;;                      unsigned panel_size,
	;;                      int u_last_adjust)

	;; A4, B4, A6, B6, A8, B8, A10, B10, A12, B12

	;; Frame Size = 112 bytes!
	
	.asg "A4",  arg_mnew
	.asg "B4",  arg_m_l1_calc
	.asg "A6",  arg_m_l1_load
	.asg "B6",  arg_l_l1
	.asg "A8",  arg_l_l2_next
	.asg "B8",  arg_u_l1
	.asg "A10", arg_m_l2_load
	.asg "B10", arg_l_l2_load
	.asg "A12", arg_u_l2
	.asg "B12", arg_panel_size
	.asg "*+SP(116)", arg_u_last_adjust 	; First arg SP+4!


	.asg "A2", mi_a
	.asg "B2", mi_b
	.asg "A3", mo_a
	.asg "B3", mo_b
	.asg "A4", l_a
	.asg "B4", l_b
	.asg "A5", u_a
	.asg "B5", u_b
	.asg "A14", panel_cnt
	.asg "A15", block_cnt
	.asg "B14", panel_reload
	.asg "B15", SP

	.asg "*+SP( 0)", u_l1_calc
	.asg "*+SP( 4)", u_l1_load
	.asg "*+SP( 8)", u_l2_load
	.asg "*+SP(12)", l_l1_load
	.asg "*+SP(16)", l_l2_load
	.asg "*+SP(20)", m_l1_load
	.asg "*+SP(24)", m_l2_load
	.asg "*+SP(28)", l_l1_calc
	.asg "*+SP(32)", l_l2_next

	.asg "0x01820100", idma1_base_addr
	.asg "0x00000200", m_load_cnt
	.asg "0xFFFFC000", u_block_adjust

update_block_l2:
	.asmfunc
	;; Layout for matrices:
	;; M 16x16 not blocked...
	
	;; Setup stack frame and save stuff
	;; SP MUST ALWAYS be aligned to 16-byte
	;; We must save: A10-A15, B10-B14, B3 (RP)

		subaw	SP, 28, SP 			; Make frame 112 bytes
	
		stdw	A11:A10, *+SP(64) 		; Registers saved 64+
		stdw	A13:A12, *+SP(72)
		stdw	A15:A14, *+SP(80)
		stdw	B11:B10, *+SP(88)
		stdw	B13:B12, *+SP(96)
		stw	B14, *+SP(104)
		stw	B3, *+SP(108)

	;; Init the local variables
		stw	arg_u_l1, u_l1_calc
		shl	arg_panel_size, 5, A3 		; U size = 32*panel_sz
		add	A3, arg_u_l1, A1
		stw	A1, u_l1_load 			; u_l1_load+=4*panel_sz
		add	arg_u_l2, A3, A2		; Load next panel
		stw	A2, u_l2_load
		add	arg_l_l1, A3, A2   	   	
		add	A2, A3, A2 			; Place to load L
		add	A2, A3, A5			; Load next half of L
		stw	A5, l_l1_load
		add	arg_l_l2_load, A3, A5 		; Also adjust L2 part
		stw	A5, l_l2_load
		stw	arg_m_l1_load, m_l1_load 	; Save m stuff for later
		stw	arg_m_l2_load, m_l2_load
		stw	arg_l_l1, l_l1_calc 		; Need for pointer swap
		stw	arg_l_l2_next, l_l2_next	; Next L panel pointer

	;; Init the first round of IDMA transfers
		mfence	
		mvkl	idma1_base_addr, A14 		; Setup a ptr to mem
		mvkh	idma1_base_addr, A14
first_idma_wait:
		ldw	*+A14[0], A0
		nop	4
	[A0]	bnop	first_idma_wait, 5

	;; Load the next U panel
		mvkh	0xe0000000, A3 			; Fill in prio=7
		stw	arg_u_l2, *+A14(8)		; Src: u_l2
		stw	A1, *+A14(12)			; Dst: u_l1_load
		stw	A3, *+A14(16)			; Cnt: panel_size*32
	;; Load 1/2 of the next L panel
		stw	arg_l_l2_load, *+A14(8) 	; Src: l_l2_load
		stw	A2, *+A14(12)			; Dst: l_l1_load
		stw	A3, *+A14(16)			; Cnt: panel_size*32

	;; Setup for the first iteration

		mvk	15, block_cnt 			; For now 16 blocks only
		add	arg_panel_size, -5, panel_cnt
		mv	arg_panel_size, panel_reload

	;; Setup pointers to input matrix (mi_a = mold, mi_b = mold + 1) 
		mv	arg_m_l1_calc, mi_a
		add	arg_m_l1_calc, 8, mi_b

	;; Setup pointers to output matrix (mo_a = mnew, mo_b = mnew + 1)
		mv	arg_mnew, mo_a
		add	arg_mnew, 8, mo_b
	
	;; Setup pointers to l panel (l_a = arg_l, l_b = arg_l + 1)
		mv	arg_l_l1, l_a
		add	arg_l_l1, 8, l_b
	
	;; Setup pointers to u panel (u_a = arg_u, u_b = arg_u + 1)
		mv	arg_u_l1, u_a
		add	arg_u_l1, 8, u_b
	

panel_restart:	
	;; Dummy setup for m

		lddw	*mi_a++[2], A17:A16 		; a00=mi
	||	lddw	*mi_b++[2], B17:B16		; a01=mi

		lddw	*mi_a++[14], A19:A18 		; a02=mi
	||	lddw	*mi_b++[14], B19:B18		; a03=mi

		lddw	*mi_a++[2], A21:A20 		; a10=mi
	||	lddw	*mi_b++[2], B21:B20		; a11=mi

		lddw	*mi_a++[14], A23:A22 		; a12=mi
	||	lddw	*mi_b++[14], B23:B22		; a13=mi

		lddw	*mi_a++[2], A25:A24 		; a20=mi
	||	lddw	*mi_b++[2], B25:B24		; a21=mi

		lddw	*mi_a++[14], A27:A26 		; a22=mi
	||	lddw	*mi_b++[14], B27:B26		; a23=mi

		lddw	*mi_a++[2], A29:A28 		; a30=mi
	||	lddw	*mi_b++[2], B29:B28		; a31=mi

		lddw	*mi_a++[14], A31:A30 		; a32=mi
	||	lddw	*mi_b++[14], B31:B30		; a33=mi

		

	;; Iteration 0
		lddw	*l_a++[2], A7:A6		; l0=l[0]
	||	lddw	*l_b++[2], B7:B6		; l1=l[0]

		lddw	*u_a++[2], A9:A8		; u0=u[0]
	||	lddw	*u_b++[2], B9:B8		; u1=u[0]

		lddw	*u_a++[2], A11:A10		; u2=u[0]
	||	lddw	*u_b++[2], B11:B10		; u3=u[0]
	
		lddw	*l_a++[2], A13:A12		; l2=l[0]
	||	lddw	*l_b++[2], B13:B12		; l3=l[0]

		nop

		zero	A1:A0				; Need two 0s
	||	zero	B1:B0
	
	;; Iteration 1

		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	lddw	*l_a++[2], A7:A6		; l0=l[1]
	||	lddw	*l_b++[2], B7:B6		; l1=l[1]

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A1:A0, A31:A30		; a32-=0 (!)
	||	fsubdp  B31:B30, B1:B0, B31:B30		; a33-=0 (!)
	||	lddw	*u_a++[2], A9:A8		; u0=u[1]
	||	lddw	*u_b++[2], B9:B8		; u1=u[1]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[1]
	||	lddw	*u_b++[2], B11:B10		; u3=u[1]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[1]
	||	lddw	*l_b++[2], B13:B12		; l3=l[1]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13

	;; Dot product loop iteration, pipelined
panel_loop:	
		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||	bdec	panel_loop, panel_cnt

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13

	;; End of panel_loop

epilogue_start:	
	;; Epilogue iteration N-1
	;; Iteration N-1 of panel_loop

		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	and	1, block_cnt, A0		; block_cnt even?

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	|| [!A0] b	next_column			; if A0==0

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13

	;; Branch taken to next_column
	
	;; Epilogue iteration 0

		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||	shl	panel_reload, 5, A0		; ud_a=pr*32
	||	shl	panel_reload, 5, B0		; ud_b=pr*32
	||	add	panel_reload, -9, panel_cnt	; pc=pr-8

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	sub	u_a, A0, u_a			; u_a-=ud_a
	||	sub	u_b, B0, u_b			; u_b-=ud_b
	||	add	block_cnt, -1, block_cnt	; bc=bc-1

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]
	
		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13
	||	stdw	A17:A16, *mo_a++[2] 		; mo=a00
	||	stdw	B17:B16, *mo_b++[2]		; mo=a01
	
	;; Epilogue iteration 1

		fmpydp	A9:A8, A7:A6, A17:A16		; i00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B17:B16		; i01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||	stdw	A19:A18, *mo_a++[14] 		; mo=a02
	||	stdw	B19:B18, *mo_b++[14]		; mo=a03

		fmpydp	A11:A10, A7:A6, A19:A18 	; i02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B19:B18		; i03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	stdw	A21:A20, *mo_a++[2] 		; mo=a10
	||	stdw	B21:B20, *mo_b++[2]		; mo=a11

		fmpydp	A9:A8, B7:B6, A21:A20		; i10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B21:B20		; i11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]

		fmpydp	A11:A10, B7:B6, A23:A22		; i12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B23:B22		; i13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A25:A24		; i20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B25:B24		; i21=l2*u1
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A27:A26	; i22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B27:B26	; i23=l2*u3
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A29:A28		; i30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B29:B28		; i31=l3*u1
	||	stdw	A23:A22, *mo_a++[14]		; mo=a12
	||	stdw	B23:B22, *mo_b++[14]		; mo=a13

		fmpydp	A11:A10, B13:B12, A1:A0		; i32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B1:B0		; i33=l3*u3
	||	stdw	A25:A24, *mo_a++[2]		; mo=a20
	||	stdw	B25:B24, *mo_b++[2]		; mo=a21

	;; Epilogue iteration 2
	
		fmpydp	A9:A8, A7:A6, A7:A6		; m00=l0*u0 (!)
	||	fmpydp	B9:B8, A7:A6, B7:B6		; m01=l0*u1
	||	stdw	A27:A26, *mo_a++[14]		; mo=a22
	||	stdw	B27:B26, *mo_b++[14]		; mo=a23

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	stdw	A29:A28, *mo_a++[2]		; mo=a30
	||	stdw	B29:B28, *mo_b++[2]		; mo=a31

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fadddp  A17:A16, A7:A6, A17:A16		; i00+=m00
	||	fadddp	B17:B16, B7:B6, B17:B16		; i01+=m01
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fadddp	A19:A18, A17:A16, A19:A18	; i02+=m02
	||	fadddp	B19:B18, B17:B16, B19:B18	; i03+=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fadddp	A21:A20, A19:A18, A21:A20	; i10+=m10
	||	fadddp	B21:B20, B19:B18, B21:B20	; i11+=m11
	||	stdw	A31:A30, *mo_a++[14]		; mo=a32
	||	stdw	B31:B30, *mo_b++[14]		; mo=a33

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fadddp	A23:A22, A21:A20, A23:A22	; i12+=m12
	||	fadddp  B23:B22, B21:B20, B23:B22	; i13+=m13
	||	lddw	*mi_a++[2], A1:A0		; a00N=mi
	||	lddw	*mi_b++[2], B1:B0		; a01N=mi

	;; Epilogue iteration 3
	
		fmpydp	A9:A8, A7:A6, A7:A6		; m00=l0*u0 (!)
	||	fmpydp	B9:B8, A7:A6, B7:B6		; m01=l0*u1 (!)
	||	fadddp	A25:A24, A23:A22, A25:A24	; i20+=m20
	||	fadddp  B25:B24, B23:B22, B25:B24	; i21+=m21
	||	lddw	*mi_a++[14], A17:A16		; a02N=mi
	||	lddw	*mi_b++[14], B17:B16		; a03N=mi

		fmpydp	A11:A10, A7:A6, A7:A6 		; m02=l0*u2 (!)
	||	fmpydp	B11:B10, A7:A6, B7:B6		; m03=l0*u3 (!)
	||	fadddp	A27:A26, A25:A24, A27:A26	; i22+=m22
	||	fadddp	B27:B26, B25:B24, B27:B26	; i23+=m23
	||	lddw	*mi_a++[2], A19:A18		; a10N=mi
	||	lddw	*mi_b++[2], B19:B18		; a11N=mi

		fmpydp	A9:A8, B7:B6, A7:A6		; m10=l1*u0 (!)
	||	fmpydp	B9:B8, B7:B6, B7:B6		; m11=l1*u1 (!)
	||	fadddp	A29:A28, A27:A26, A29:A28	; i30+=m30
	||	fadddp	B29:B28, B27:B26, B29:B28	; i31+=m31
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2 (!)
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3 (!)
	||	fadddp	A1:A0, A29:A28, A31:A30		; i32+=m32 (!)
	||	fadddp	B1:B0, B29:B28, B31:B30		; i33+=m33 (!)
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A21:A20		; m20=l2*u0 (!)
	||	fmpydp  B9:B8, A13:A12, B21:B20		; m21=l2*u1 (!)
	||	fadddp  A17:A16, A7:A6, A9:A8		; i00+=m00 (!)
	||	fadddp	B17:B16, B7:B6, B9:B8		; i01+=m01 (!)
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A21:A20	; m22=l2*u2 (!)
	||	fmpydp	B11:B10, A13:A12, B21:B20	; m23=l2*u3 (!)
	||	fadddp	A19:A18, A7:A6, A11:A10		; i02+=m02 (!)
	||	fadddp	B19:B18, B7:B6, B11:B10		; i03+=m03 (!)
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A21:A20		; m30=l3*u0 (!)
	||	fmpydp	B9:B8, B13:B12, B21:B20		; m31=l3*u1 (!)
	||	fadddp	A21:A20, A7:A6, A17:A16		; i10+=m10 (!)
	||	fadddp	B21:B20, B7:B6, B17:B16		; i11+=m11 (!)
	||	lddw	*mi_a++[14], A21:A20		; a12N=mi
	||	lddw	*mi_b++[14], B21:B20		; a13N=mi

		fmpydp	A11:A10, B13:B12, A23:A22	; m32=l3*u2 (!)
	||	fmpydp	B11:B10, B13:B12, B23:B22	; m33=l3*u3 (!)
	||	fadddp	A23:A22, A21:A20, A27:A26	; i12+=m12 (!)
	||	fadddp  B23:B22, B21:B20, B27:B26	; i13+=m13 (!)
	||	fsubdp	A1:A0, A9:A8, A17:A16		; a00-=i00
	||	fsubdp	B1:B0, B9:B8, B17:B16		; a01-=i01
	||	lddw	*mi_a++[2], A7:A6		; a20N=mi
	||	lddw	*mi_b++[2], B7:B6		; a21N=mi

	;; Epilogue iteration 4
	
		fmpydp	A9:A8, A7:A6, A23:A22		; m00=l0*u0 (!)
	||	fmpydp	B9:B8, A7:A6, B23:B22		; m01=l0*u1 (!)
	||	fadddp	A25:A24, A21:A20, A29:A28	; i20+=m20 (!)
	||	fadddp  B25:B24, B21:B20, B29:B28	; i21+=m21 (!)
	||	fsubdp	A17:A16, A11:A10, A19:A18	; a02-=i02
	||	fsubdp	B17:B16, B11:B10, B19:B18	; a03-=i03
	||	lddw	*mi_a++[14], A7:A6		; a22N=mi
	||	lddw	*mi_b++[14], B7:B6		; a23N=mi

		fmpydp	A11:A10, A7:A6, A23:A22 	; m02=l0*u2 (!)
	||	fmpydp	B11:B10, A7:A6, B23:B22		; m03=l0*u3 (!)
	||	fadddp	A27:A26, A21:A20, A31:A30	; i22+=m22 (!)
	||	fadddp	B27:B26, B21:B20, B31:B30	; i23+=m23 (!)
	||	fsubdp	A19:A18, A17:A16, A21:A20	; a10-=i10
	||	fsubdp	B19:B18, B17:B16, B21:B20	; a11-=i11
	||	lddw	*mi_a++[2], A7:A6		; a30N=mi
	||	lddw	*mi_b++[2], B7:B6		; a31N=mi

		fmpydp	A9:A8, B7:B6, A25:A24		; m10=l1*u0 (!)
	||	fmpydp	B9:B8, B7:B6, B25:B24		; m11=l1*u1 (!)
	||	fadddp	A29:A28, A21:A20, A17:A16	; i30+=m30 (!)
	||	fadddp	B29:B28, B21:B20, B17:B16	; i31+=m31 (!)
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]

		fmpydp	A11:A10, B7:B6, A27:A26		; m12=l1*u2 (!)
	||	fmpydp	B11:B10, B7:B6, B27:B26		; m13=l1*u3 (!)
	||	fadddp	A31:A30, A23:A22, A31:A30	; i32+=m32 (!)
	||	fadddp	B31:B30, B23:B22, B31:B30	; i33+=m33 (!)
	||	fsubdp	A21:A20, A27:A26, A23:A22	; a12-=i12
	||	fsubdp	B21:B20, B27:B26, B23:B22	; a13-=i13
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A29:A28		; m20=l2*u0 (!)
	||	fmpydp  B9:B8, A13:A12, B29:B28		; m21=l2*u1 (!)
	||	fsubdp  A17:A16, A23:A22, A17:A16	; a00-=m00 (!)
	||	fsubdp	B17:B16, B23:B22, B17:B16	; a01-=m01 (!)
	||	fsubdp	A7:A6, A29:A28, A25:A24		; a20-=i20
	||	fsubdp	B7:B6, B29:B28, B25:B24		; a21-=i21
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A23:A22, A19:A18	; a02-=m02 (!)
	||	fsubdp	B19:B18, B23:B22, B19:B18	; a03-=m03 (!)
	||	fsubdp	A7:A6, A31:A30, A27:A26		; a22-=i22
	||	fsubdp	B7:B6, B31:B30, B27:B26		; a23-=i23
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0 
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1 
	||	fsubdp	A21:A20, A25:A24, A21:A20	; a10-=m10 (!)
	||	fsubdp	B21:B20, B25:B24, B21:B20	; a11-=m11 (!)
	||	fsubdp	A7:A6, A17:A16, A29:A28		; a30-=i30
	||	fsubdp	B7:B6, B17:B16, B29:B28		; a31-=i31
	||	lddw	*mi_a++[14], A27:A26		; a32N=mi
	||	lddw	*mi_b++[14], B27:B26		; a33N=mi

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2 
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3 
	||	fsubdp	A23:A22, A27:A26, A23:A22	; a12-=m12 (!)
	||	fsubdp  B23:B22, B27:B26, B23:B22	; a13-=m13 (!)

	;; Epilogue iteration 5
	
		fmpydp	A9:A8, A7:A6, A29:A28		; m00=l0*u0 (!)
	||	fmpydp	B9:B8, A7:A6, B29:B28		; m01=l0*u1 (!)
	||	fsubdp	A25:A24, A29:A28, A25:A24	; a20-=m20 (!)
	||	fsubdp  B25:B24, B29:B28, B25:B24	; a21-=m21 (!)

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22 
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23 

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30 
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31 
	||	lddw	*l_a++[2], A7:A6		; l0=l[i]
	||	lddw	*l_b++[2], B7:B6		; l1=l[i]
	||	b	panel_loop

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A27:A26, A29:A28, A17:A16	; a32-=m32 (!N)
	||	fsubdp	B27:B26, B29:B28, B17:B16	; a33-=m33 (!N)
	||	lddw	*u_a++[2], A9:A8		; u0=u[i]
	||	lddw	*u_b++[2], B9:B8		; u1=u[i]

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A29:A28, A17:A16	; a00-=m00 (!)
	||	fsubdp	B17:B16, B29:B28, B17:B16	; a01-=m01 (!)
	||	lddw	*u_a++[2], A11:A10		; u2=u[i]
	||	lddw	*u_b++[2], B11:B10		; u3=u[i]

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03
	||	lddw	*l_a++[2], A13:A12		; l2=l[i]
	||	lddw	*l_b++[2], B13:B12		; l3=l[i]

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11
	||	fsubdp	A17:A16, A31:A30, A31:A30	; a32-=i32
	||	fsubdp	B17:B16, B31:B30, B31:B30	; a33-=i33

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13

	;; Branch taken to panel_loop

next_column:
	
	;; We have to start working on a new column
	;; For now drain the pipe and write back old block
	;; In the Pipe: L(N-1) M(N-2) A(N-3)

	;; Drain 1: M(N-1) A(N-2)

		fmpydp	A9:A8, A7:A6, A31:A30		; m00=l0*u0
	||	fmpydp	B9:B8, A7:A6, B31:B30		; m01=l0*u1
	||	fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21

		fmpydp	A11:A10, A7:A6, A17:A16 	; m02=l0*u2
	||	fmpydp	B11:B10, A7:A6, B17:B16		; m03=l0*u3
	||	fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23

		fmpydp	A9:A8, B7:B6, A19:A18		; m10=l1*u0
	||	fmpydp	B9:B8, B7:B6, B19:B18		; m11=l1*u1
	||	fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31

		fmpydp	A11:A10, B7:B6, A21:A20		; m12=l1*u2
	||	fmpydp	B11:B10, B7:B6, B21:B20		; m13=l1*u3
	||	fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33

		fmpydp	A9:A8, A13:A12, A23:A22		; m20=l2*u0
	||	fmpydp  B9:B8, A13:A12, B23:B22		; m21=l2*u1
	||	fsubdp  A17:A16, A31:A30, A17:A16	; a00-=m00
	||	fsubdp	B17:B16, B31:B30, B17:B16	; a01-=m01

		fmpydp	A11:A10, A13:A12, A25:A24	; m22=l2*u2
	||	fmpydp	B11:B10, A13:A12, B25:B24	; m23=l2*u3
	||	fsubdp	A19:A18, A17:A16, A19:A18	; a02-=m02
	||	fsubdp	B19:B18, B17:B16, B19:B18	; a03-=m03

		fmpydp	A9:A8, B13:B12, A27:A26		; m30=l3*u0
	||	fmpydp	B9:B8, B13:B12, B27:B26		; m31=l3*u1
	||	fsubdp	A21:A20, A19:A18, A21:A20	; a10-=m10
	||	fsubdp	B21:B20, B19:B18, B21:B20	; a11-=m11

		fmpydp	A11:A10, B13:B12, A29:A28	; m32=l3*u2
	||	fmpydp	B11:B10, B13:B12, B29:B28	; m33=l3*u3
	||	fsubdp	A23:A22, A21:A20, A23:A22	; a12-=m12
	||	fsubdp  B23:B22, B21:B20, B23:B22	; a13-=m13
	
	;; Drain 2: A(N-3)

		fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21

		fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	
		fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31

		fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	
	;; Drain 3: S(Mnew)

		stdw	A17:A16, *mo_a++[2] 		; mo=a00
	||	stdw	B17:B16, *mo_b++[2]		; mo=a01

		stdw	A19:A18, *mo_a++[14] 		; mo=a02
	||	stdw	B19:B18, *mo_b++[14]		; mo=a03

		stdw	A21:A20, *mo_a++[2] 		; mo=a10
	||	stdw	B21:B20, *mo_b++[2]		; mo=a11

		stdw	A23:A22, *mo_a++[14] 		; mo=a12
	||	stdw	B23:B22, *mo_b++[14]		; mo=a13

		stdw	A25:A24, *mo_a++[2] 		; mo=a20
	||	stdw	B25:B24, *mo_b++[2]		; mo=a21

		stdw	A27:A26, *mo_a++[14] 		; mo=a22
	||	stdw	B27:B26, *mo_b++[14]		; mo=a23

		stdw	A29:A28, *mo_a++[2] 		; mo=a30
	||	stdw	B29:B28, *mo_b++[2]		; mo=a31

		stdw	A31:A30, *mo_a++[14] 		; mo=a32
	||	stdw	B31:B30, *mo_b++[14]		; mo=a33

	;; Now do the IDMA and pointer updates
		mv	block_cnt, A0 			; bc==0
	[!A0]	b	cleanup				; Return to caller
		nop	5				; Delay slots
	
idma_setup:	
	;; Check that the IDMA is done with any xfers
		mvkl	idma1_base_addr, A14 		; Setup a ptr to mem
		mvkh	idma1_base_addr, A14
idma_wait:
		ldw	*+A14[0], A0 			; load status
		nop	4
	[A0]	b	idma_wait 			; 0 indicates done
		nop	5
	;; Branch to idma_wait occurrs

		ldw	u_l1_load, u_a 		; Swap pointers
		ldw	u_l1_calc, A6
		ldw	u_l2_load, A7
		ldw	arg_u_last_adjust, A9
		ldw	l_l1_load, A10
		ldw	l_l2_load, A11
		ldw	m_l1_load, A12
		ldw	m_l2_load, A13
		nop	4				; FIXME!
	
		shl	panel_reload, 5, A1 		; A1: panel_size*32
		sub	block_cnt, 2, A0		; block_cnt == 2
	[!A0]	add	A7, A9, A7			; if last, adjust U
		sub	block_cnt, 10, A0		; block_cnt == 10
		mvk	u_block_adjust, A9		; Next row start
	[!A0]	add	A7, A9, A7
		add	A1, A7, A8
		stw	A8, u_l2_load 			; Next panel start
		sub	block_cnt, 8, A0		; block_cnt == 8
	[!A0]	ldw	l_l1_calc, A10 			; Reset L dst pointer
	[!A0]	ldw	l_l2_next, A11			; Reset L src pointer
		stw	A6, u_l1_load
		stw	u_a, u_l1_calc
		add	u_a, 8, u_b 			; U pointers set up
		mv	A1, A17				; Stash away for later
		mvkh	0xe0000000, A1			; Set IDMA priority 7
		stw	A7, *+A14(8)			; Src: u_l2_load
		stw	A6, *+A14(12)			; Dst: u_l1_load
		stw	A1, *+A14(16) 			; Cnt: panel_size*32

	;; Update normal L/M pointers pointers
	[A0]	mvk	0xfffffc20, A6		    	; Move M ptr back...
	[!A0]	mvk	0xffffffa0, A6			; Block 8, 12 columns!
		add	mi_a, A6, mi_a
		add	mo_a, A6, mo_a
		add	mi_b, A6, mi_b
		add	mo_b, A6, mo_b
		shl	panel_reload, 6, A6	    	; Also need to update L
	[A0]	sub	l_a, A6, l_a
	[A0]	sub	l_b, A6, l_b
	
		;; sub	block_cnt, 6, A0 		; block_cnt == 6
		and	block_cnt, 4, A0 		; block_cnt & 4 
	[!A0]	stw	A11, *+A14(8)			; Src: l_l2_load
	[!A0]	stw	A10, *+A14(12)			; Dst: l_l1_load
	[!A0]	stw	A1, *+A14(16)			; Cnt: panel_size*32
	[!A0]	add	A10, A17, A10			; Increment L pointer
	[!A0]	stw	A10, l_l1_load			; Update L src pointer
	[!A0]	add	A11, A17, A11			; Increment L pointer
	[!A0]	stw	A11, l_l2_load			; Update L dst pointer

		mvk	m_load_cnt, A1 			; Bring in the count
		add	A1, A13, A7
	[A0]	stw	A7, m_l2_load 			; m_l2_load+=cnt
		add	A1, A12, A6
	[A0]	stw	A6, m_l1_load 			; m_l1_load+=cnt
		mvkh	0xe0000000, A1			; Set IDMA priority 7
	[A0]	stw	A13, *+A14(8)			; Src: m_l2_load
	[A0]	stw	A12, *+A14(12)			; Dst: m_l1_load
	[A0]	stw	A1, *+A14(16)			; Cnt: block_size*32
		
		
restart_iter:	
	
	;; Restart iteration
		add	panel_reload, -5, panel_cnt	; pc=pr-5
		add	block_cnt, -1, block_cnt 	; bc=bc-1
		b	panel_restart
		nop	5				; Delay slots


exit_epilogue:
	;; Exit iteration 1
		fsubdp	A25:A24, A23:A22, A25:A24	; a20-=m20
	||	fsubdp  B25:B24, B23:B22, B25:B24	; a21-=m21
	||	stdw	A19:A18, *mo_a++[14] 		; mo=a02
	||	stdw	B19:B18, *mo_b++[14]		; mo=a03

		fsubdp	A27:A26, A25:A24, A27:A26	; a22-=m22
	||	fsubdp	B27:B26, B25:B24, B27:B26	; a23-=m23
	||	stdw	A21:A20, *mo_a++[2] 		; mo=a10
	||	stdw	B21:B20, *mo_b++[2]		; mo=a11

		fsubdp	A29:A28, A27:A26, A29:A28	; a30-=m30
	||	fsubdp	B29:B28, B27:B26, B29:B28	; a31-=m31
	||	stdw	A23:A22, *mo_a++[14] 		; mo=a12
	||	stdw	B23:B22, *mo_b++[14]		; mo=a13

		fsubdp	A31:A30, A29:A28, A31:A30	; a32-=m32
	||	fsubdp	B31:B30, B29:B28, B31:B30	; a33-=m33
	||	stdw	A25:A24, *mo_a++[2] 		; mo=a20
	||	stdw	B25:B24, *mo_b++[2]		; mo=a21

		stdw	A27:A26, *mo_a++[14] 		; mo=a22
	||	stdw	B27:B26, *mo_b++[14]		; mo=a23

		stdw	A29:A28, *mo_a++[2] 		; mo=a30
	||	stdw	B29:B28, *mo_b++[2]		; mo=a31

		stdw	A31:A30, *mo_a++[14] 		; mo=a32
	||	stdw	B31:B30, *mo_b++[14]		; mo=a33
	
	;; Clean-up & restore registers

cleanup:
		ldw	*+SP(108), B3
		ldw	*+SP(104), B14
		lddw	*+SP(96), B13:B12
		lddw	*+SP(88), B11:B10
		lddw	*+SP(80), A15:A14
		lddw	*+SP(72), A13:A12
		lddw	*+SP(64), A11:A10
		addaw	SP, 28, SP 			; Frame was 112 bytes
		bnop	B3, 5
	;; Return from function here
	.endasmfunc