/**
   Simple DMA based Linpack benchmark.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "dma-linpack.h"

void
update_u_panel_l2(double * restrict u_l2,
		  const double * restrict l_head,
		  size_t panel_width,
		  size_t cols)
{
  double sum_00, sum_01, sum_02, sum_03;
  double sum_10, sum_11, sum_12, sum_13;
  double sum_20, sum_21, sum_22, sum_23;
  double sum_30, sum_31, sum_32, sum_33;

  double l_0, l_1, l_2, l_3;
  double u_0, u_1, u_2, u_3;

  double * restrict u;
  const double * restrict l;

  size_t row;
  size_t col_block;
  size_t col_ofs;
  size_t r;

  for(col_block=0; col_block<cols; col_block+=BLOCK_SIZE) {
    for(row=0; row<panel_width; row+=SUBBLOCK_SIZE) {
      for(col_ofs=0; col_ofs<BLOCK_SIZE; col_ofs += SUBBLOCK_SIZE) {

	u = u_l2 + (col_ofs + col_block)*panel_width;
	l = l_head + row*panel_width;

	sum_00 = 0.0; sum_01 = 0.0; sum_02 = 0.0; sum_03 = 0.0;
	sum_10 = 0.0; sum_11 = 0.0; sum_12 = 0.0; sum_13 = 0.0;
	sum_20 = 0.0; sum_21 = 0.0; sum_22 = 0.0; sum_23 = 0.0;
	sum_30 = 0.0; sum_31 = 0.0; sum_32 = 0.0; sum_33 = 0.0;

	for(r=row; r>0; --r) {
	  u_0 = u[0]; u_1 = u[1]; u_2 = u[2]; u_3 = u[3];
	  l_0 = l[0]; l_1 = l[1]; l_2 = l[2]; l_3 = l[3];
	  u+=4; 
	  l+=4;

	  sum_00 += u_0*l_0;
	  sum_01 += u_1*l_0;
	  sum_02 += u_2*l_0;
	  sum_03 += u_3*l_0;

	  sum_10 += u_0*l_1;
	  sum_11 += u_1*l_1;
	  sum_12 += u_2*l_1;
	  sum_13 += u_3*l_1;

	  sum_20 += u_0*l_2;
	  sum_21 += u_1*l_2;
	  sum_22 += u_2*l_2;
	  sum_23 += u_3*l_2;

	  sum_30 += u_0*l_3;
	  sum_31 += u_1*l_3;
	  sum_32 += u_2*l_3;
	  sum_33 += u_3*l_3;
	}

	u_0 = u[0]; u_1 = u[1]; u_2 = u[2]; u_3 = u[3];
	l_0 = l[0]; l_1 = l[1]; l_2 = l[2]; l_3 = l[3];

	sum_00 = l_0*(u_0 - sum_00); 
	sum_01 = l_0*(u_1 - sum_01);
	sum_02 = l_0*(u_2 - sum_02);
	sum_03 = l_0*(u_3 - sum_03);

	u[0] = sum_00;
	u[1] = sum_01;
	u[2] = sum_02;
	u[3] = sum_03;

	u+=4;
	l+=4;

	sum_10 += sum_00*l_1;
	sum_11 += sum_01*l_1;
	sum_12 += sum_02*l_1;
	sum_13 += sum_03*l_1;
	
	sum_20 += sum_00*l_2;
	sum_21 += sum_01*l_2;
	sum_22 += sum_02*l_2;
	sum_23 += sum_03*l_2;

	sum_30 += sum_00*l_3;
	sum_31 += sum_01*l_3;
	sum_32 += sum_02*l_3;
	sum_33 += sum_03*l_3;

	u_0 = u[0]; u_1 = u[1]; u_2 = u[2]; u_3 = u[3];
	l_1 = l[1]; l_2 = l[2]; l_3 = l[3];

	sum_10 = l_1*(u_0 - sum_10); 
	sum_11 = l_1*(u_1 - sum_11);
	sum_12 = l_1*(u_2 - sum_12);
	sum_13 = l_1*(u_3 - sum_13);

	u[0] = sum_10;
	u[1] = sum_11;
	u[2] = sum_12;
	u[3] = sum_13;

	u+=4;
	l+=4;
	
	sum_20 += sum_10*l_2;
	sum_21 += sum_11*l_2;
	sum_22 += sum_12*l_2;
	sum_23 += sum_13*l_2;

	sum_30 += sum_10*l_3;
	sum_31 += sum_11*l_3;
	sum_32 += sum_12*l_3;
	sum_33 += sum_13*l_3;

	u_0 = u[0]; u_1 = u[1]; u_2 = u[2]; u_3 = u[3];
	l_2 = l[2]; l_3 = l[3];

	sum_20 = l_2*(u_0 - sum_20); 
	sum_21 = l_2*(u_1 - sum_21);
	sum_22 = l_2*(u_2 - sum_22);
	sum_23 = l_2*(u_3 - sum_23);

	u[0] = sum_20;
	u[1] = sum_21;
	u[2] = sum_22;
	u[3] = sum_23;

	u+=4;
	l+=4;
	
	sum_30 += sum_20*l_3;
	sum_31 += sum_21*l_3;
	sum_32 += sum_22*l_3;
	sum_33 += sum_23*l_3;

	u_0 = u[0]; u_1 = u[1]; u_2 = u[2]; u_3 = u[3];
	l_3 = l[3];

	sum_30 = l_3*(u_0 - sum_30); 
	sum_31 = l_3*(u_1 - sum_31);
	sum_32 = l_3*(u_2 - sum_32);
	sum_33 = l_3*(u_3 - sum_33);

	u[0] = sum_30;
	u[1] = sum_31;
	u[2] = sum_32;
	u[3] = sum_33;	
      }
    }
  }

#if 0
  double row_scale;
  double sum;
  size_t row;
  size_t col;
  size_t r;

  /* Eliminate equations in the upper panel, scale at the same time. */
  for(row=0; row<panel_width; ++row) {
    row_scale = l_head[index_l_l2(row, row)];

    for(col=0; col<cols; ++col) {
      sum = 0.0;
      for(r = 0; r<row; ++r) {
	sum += l_head[index_l_l2(row, r)]*u_l2[index_u_l2(r, col)];
      } 

      u_l2[index_u_l2(row, col)] = row_scale*(u_l2[index_u_l2(row, col)] - sum);
    }
  }
#endif
}
