/*
  Simple cache control.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "include/cache.h"
#include "include/corepack_regs.h"

#include <stdio.h>

#define CACHE_MAR_MASK (~((uint32_t)(CACHE_MAR_SIZE - 1)))

/*#define CACHE_MAR_BASE ((uint32_t *)(0x01848000ul))*/
#define CACHE_MAR_ATTR_MASK (CACHE_MAR_CACHEABLE | CACHE_MAR_PREFETCHABLE)

unsigned cache_set_memory_attributes(void *base, 
				     size_t size, 
				     uint32_t attributes)
{
  uint32_t start_addr;
  uint32_t end_addr;
  volatile uint32_t *mar_first, *mar_last, *mar_idx;
  uint32_t attrs;

  attrs = attributes & CACHE_MAR_ATTR_MASK;

  start_addr = ((uint32_t)(base)) & CACHE_MAR_MASK;
  end_addr = (start_addr + size) & CACHE_MAR_MASK;

  mar_first = corepack_regs->cache_regs.mar + (start_addr / CACHE_MAR_SIZE);
  mar_last  = corepack_regs->cache_regs.mar + (end_addr / CACHE_MAR_SIZE) + 1;

  for (mar_idx = mar_first; mar_idx<mar_last; mar_idx++) {
    *mar_idx = (*mar_idx & ~CACHE_MAR_ATTR_MASK ) | attrs;
  }

  return 0;
}

uint32_t cache_get_memory_attribtues(void *base)
{
  uint32_t start_addr;
  volatile uint32_t *mar_idx;

  start_addr = ((uint32_t)(base)) & CACHE_MAR_MASK;
  mar_idx = corepack_regs->cache_regs.mar + (start_addr / CACHE_MAR_SIZE);

  return *mar_idx;
}

unsigned
cache_l2_set_size(uint32_t size_log_2)
{
  uint32_t config_mode;

  if (size_log_2 > 7) return 1;

  config_mode = corepack_regs->cache_regs.l2cfg;
  config_mode = (config_mode & 0xfffffff8ul) | (size_log_2 & 0x7ul);
  corepack_regs->cache_regs.l2cfg = config_mode;

  config_mode = corepack_regs->cache_regs.l2cfg;
  if ((config_mode & 0x7ul) != size_log_2) return 1;
  return 0;
}

unsigned
cache_l1d_set_size(uint32_t size_log_2)
{
  uint32_t config_mode;

  if (size_log_2 > 7) return 1;

  config_mode = corepack_regs->cache_regs.l1dcfg;
  config_mode = (config_mode & 0xfffffff8ul) | (size_log_2 & 0x7ul);
  corepack_regs->cache_regs.l1dcfg = config_mode;

  config_mode = corepack_regs->cache_regs.l1dcfg;
  if ((config_mode & 0x7ul) != size_log_2) return 1;
  return 0;
}
