/**
   EDMA3 registers.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#ifndef RT_EDMA3_REGS_H
#define RT_EDMA3_REGS_H

#include <stdint.h>

/* 
   A PaRAM Set Entry, needs to be accessed as 32-bit words.
 */
struct edma3_param_set {
  uint32_t opt;			/* 0x00 */
  uint32_t src;			/* 0x04 */
  uint32_t abcnt;		/* 0x08 */
  uint32_t dst;			/* 0x0c */
  uint32_t bidx;		/* 0x10 */
  uint32_t link_bcntrld;	/* 0x14 */
  uint32_t cidx;		/* 0x18 */
  uint32_t ccnt;		/* 0x1c */
};

/*
  EDMA3 Channel Controller DMA Region Access Enable registers.
*/
struct edma3cc_drae_registers {
  uint32_t low;			/* 0x0 */
  uint32_t high;		/* 0x8 */
};

/*
  EDMA3 Channel Controller Channel registers.

  Size should be 0x200 bytes.
*/
struct edma3cc_channel_registers {
  uint32_t er;			/* 0x000 */
  uint32_t erh;			/* 0x004 */
  uint32_t ecr;			/* 0x008 */
  uint32_t ecrh;		/* 0x00c */
  uint32_t esr;			/* 0x010 */
  uint32_t esrh;		/* 0x014 */
  uint32_t cer;			/* 0x018 */
  uint32_t cerh;		/* 0x01c */
  uint32_t eer;			/* 0x020 */
  uint32_t eerh;		/* 0x024 */
  uint32_t eecr;		/* 0x028 */
  uint32_t eecrh;		/* 0x02c */
  uint32_t eesr;		/* 0x030 */
  uint32_t eesrh;		/* 0x034 */
  uint32_t ser;			/* 0x038 */
  uint32_t serh;		/* 0x03c */
  uint32_t secr;		/* 0x040 */
  uint32_t secrh;		/* 0x044 */
  uint32_t res1[2];		/* 0x048 - 0x04f */
  uint32_t ier;			/* 0x050 */
  uint32_t ierh;		/* 0x054 */
  uint32_t iecr;		/* 0x058 */
  uint32_t iecrh;		/* 0x05c */
  uint32_t iesr;		/* 0x060 */
  uint32_t iesrh;		/* 0x064 */
  uint32_t ipr;			/* 0x068 */
  uint32_t iprh;		/* 0x06c */
  uint32_t icr;			/* 0x070 */
  uint32_t icrh;		/* 0x074 */
  uint32_t ieval;		/* 0x078 */
  uint32_t res2;		/* 0x07c */
  uint32_t qer;			/* 0x080 */
  uint32_t qeer;		/* 0x084 */
  uint32_t qeecr;		/* 0x088 */
  uint32_t qeesr;		/* 0x08c */
  uint32_t qser;		/* 0x090 */
  uint32_t qsecr;		/* 0x094 */
  uint32_t res3[90];		/* 0x098 - 0x1ff */
};

/*
  EDMA3 Channel Controller Registers.

  Each channel controller occupies 0x20000 bytes of address space. 
*/
struct edma3cc_registers {
  uint32_t pid;			/* 0x0000 */
  uint32_t cccfg;		/* 0x0004 */
  uint32_t res1[62];		/* 0x0008 - 0x00ff */
  uint32_t dchmap[64];		/* 0x0100 - 0x01ff */
  uint32_t qchmap[8];		/* 0x0200 - 0x021f */
  uint32_t res2[8];		/* 0x0220 - 0x023f */
  uint32_t dmaqnum[8];		/* 0x0240 - 0x025f */
  uint32_t qdmaqnum;		/* 0x0260 */
  uint32_t res3[7];		/* 0x0264 - 0x027f */
  uint32_t quetcmap;		/* 0x0280 */
  uint32_t quepri;		/* 0x0284 */
  uint32_t res4[30];		/* 0x0288 - 0x02ff */
  uint32_t emr;			/* 0x0300 */
  uint32_t emrh;		/* 0x0304 */
  uint32_t emcr;		/* 0x0308 */
  uint32_t emcrh;		/* 0x030c */
  uint32_t qemr;		/* 0x0310 */
  uint32_t qemcr;		/* 0x0314 */
  uint32_t ccerr;		/* 0x0318 */
  uint32_t ccerrclr;		/* 0x031c */
  uint32_t eeval;		/* 0x0320 */
  uint32_t res5[7];		/* 0x0324 - 0x033f */
  struct edma3cc_drae_registers drae[8]; /* 0x0340 - 0x037f */
  uint32_t qrae[8];		/* 0x0380 - 0x039f */
  uint32_t res6[24];		/* 0x03a0 - 0x03ff */
  uint32_t qxey[64];		/* 0x0400 - 0x04ff */
  uint32_t res7[704];		/* 0x0500 - 0x0fff */
  struct edma3cc_channel_registers global; /* 0x1000 - 0x11ff */
  uint32_t res8[896];		/* 0x1200 - 0x1fff */
  struct edma3cc_channel_registers shadow[8]; /* 0x2000 - 0x2fff */
  uint32_t res9[1024];		/* 0x3000 - 0x3fff */
  struct edma3_param_set param[512]; /* 0x4000 - 0x7fff */
  uint32_t res10[24576];	/* 0x8000 - 0x1ffff */
};

/*
  EDMA3 Transfer Controller Registers.

  Each transfer controller occupies 0x08000 bytes of address space.
*/
struct edma3tc_registers {
  uint32_t res1[8192];		/* 0x0000 - 0x7ffff */
};

/* EDMA3 peripherials in C6678. */

static volatile struct edma3cc_registers * const edma3cc_registers =
  (struct edma3cc_registers *)(0x02700000ul);

static volatile struct edma3tc_registers * const edma3tc_registers =
  (struct edma3tc_registers *)(0x02760000ul);


#endif
