/**
   HyperLink runtime support functions.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#ifndef RT_HL_H
#define RT_HL_H

#include <stdint.h>

/**
   Select the local or remote information.
*/
enum hl_location {
  HL_LOCATION_LOCAL,
  HL_LOCATION_REMOTE
};

/**
   Initialize the HyperLink connection.

   \args   param_idx  Index into link parameter table (default: 0)
   \return 0 if successful, errno otherwise
*/  
int hl_init(unsigned param_idx);

static const unsigned HL_PARAM_DEFAULT=0;

/**
   Get the revision id of the HyperLink module.

   \return The 32-bit revision ID from the module register.
*/
uint32_t hl_get_revision(enum hl_location location);

/**
   Get the current error counters and maybe reset them.

   \args   location   Which counters (local, remote) to read.
   \args   reset      if not 0 reset the counters to 0.
   \args   cnt_single counter of (corrected) single bit errors.
   \args   cnt_double counter of (detected) double bit errors.
   \return            0 if successful, errno otherwise.
*/
int hl_get_error_counters(enum hl_location location,
			  int reset,
			  unsigned *cnt_single,
			  unsigned *cnt_double);

/**
   Set the ingress address mapping.

   \args   remote_start  Start address, should be aligned to 4MB.
   \args   local_start   Mapped address start, aligned to 4MB.
   \args   size          Mapping size, aligend to 4MB

   \return 0 if successful, errno otherwise
*/
int hl_ingess_addr_map(uint32_t remote_start, 
                       void *local_start,
		       uint32_t size);

#endif
