/**
   HyperLink register module support.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#ifndef RT_HL_REGS_H
#define RT_HL_REGS_H

#include <stdint.h>

/**
   Control register structure for HyperLink module.

   The total size of this struct should be 128 bytes!
*/
struct hl_control_registers
{
  uint32_t revision_id;		/* 0x00 */
  uint32_t control;		/* 0x04 */
  uint32_t status;		/* 0x08 */
  uint32_t interrupt_priority;	/* 0x0c */
  uint32_t interrupt_status;	/* 0x10 */
  uint32_t interrupt_pending;	/* 0x14 */
  uint32_t interrupt_generate;	/* 0x18 */
  uint32_t tx_addr_overlay;	/* 0x1c */
  uint32_t res1[3];		/* 0x20 - 0x28 */
  uint32_t rx_addr_selector;	/* 0x2c */
  uint32_t rx_addr_privid_idx;	/* 0x30 */
  uint32_t rx_addr_privid_val;	/* 0x34 */
  uint32_t rx_addr_seg_idx;	/* 0x38 */
  uint32_t rx_addr_seg_val;	/* 0x3c */
  uint32_t chip_version;	/* 0x40 */
  uint32_t lane_power_mgmt;	/* 0x44 */
  uint32_t res2;		/* 0x48 */
  uint32_t ecc_error_counters;	/* 0x4c */
  uint32_t res3[4];		/* 0x50 - 0x5c */
  uint32_t interrupt_ctrl_idx;	/* 0x60 */
  uint32_t interrupt_ctrl_val;	/* 0x64 */
  uint32_t interrupt_ptr_idx;	/* 0x68 */
  uint32_t interrupt_ptr_val;	/* 0x6c */
  uint32_t serdes_ctrl_stat[4];	/* 0x70 - 0x7c*/
};

/**
   Register definition for HyperLink module address space.
*/
struct hl_module_registers 
{
  struct hl_control_registers local;
  struct hl_control_registers remote;
};


/*
  Hyper link module registers base.
*/
static volatile struct hl_module_registers * const hl_module_registers =
  (struct hl_module_registers *)0x21400000ul;

#endif
