/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "log.h"
#include "uart.h"

#include <stdio.h>

static char msg_buffer[LOG_LINE_LEN];

void 
log_init()
{
  uart_init();
}

void 
log_write_start()
{
  uart_puts("S\r");
}

void 
log_write_end()
{
  uart_puts("E\r");
}

void 
log_write_marker(const char *name)
{
  sprintf(msg_buffer, "T\t%s\r", name);
  uart_puts(msg_buffer);
}

void 
log_write_result(const char *name, const char *value)
{
  sprintf(msg_buffer, "R\t%s\t%s\r", name, value);
  uart_puts(msg_buffer);
}

void 
log_write_result_uint64(const char *name, const uint64_t value)
{
  sprintf(msg_buffer, "R\t%s\t%llu\r", name, value);
  uart_puts(msg_buffer);
}

void 
log_write_result_double(const char *name, const double value)
{
  sprintf(msg_buffer, "R\t%s\t%20.10g\r", name, value);
  uart_puts(msg_buffer);
}


void 
log_write_param(const char *name, const char *value)
{
  sprintf(msg_buffer, "P\t%s\t%s\r", name, value);
  uart_puts(msg_buffer);
}

void 
log_write_param_uint64(const char *name, const uint64_t value)
{
  sprintf(msg_buffer, "P\t%s\t%llu\r", name, value);
  uart_puts(msg_buffer);
}

void 
log_write_name(const char *name)
{
  sprintf(msg_buffer, "N\t%s\r", name);
  uart_puts(msg_buffer);
}

void 
log_write_message(const char *msg)
{
  sprintf(msg_buffer, "M\t%s\r", msg);
  uart_puts(msg_buffer);
}

void 
log_write(const char *tag, const char *msg)
{
  sprintf(msg_buffer, "%s\t%s\r", tag, msg);
  uart_puts(msg_buffer);
}
