/**
   Simple initialization for the platform.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include <tsc.h>
#include <cache.h>
#include <log.h>
#include <barrier.h>
#include <sem.h>

#include <_lock.h>
#include <stdlib.h>

/**
   Setup common things on the DSPs.

   This is called before main() is executed, stack is already set up.
*/
void
platform_init()
{
  int my_rank;

  tsc_init();
  log_init();

  /* Set L2 to 512K cache/ no SRAM... */
  if (cache_l2_set_size(5) != 0) {
#if 0
    printf("Could not set up L2 cache!\n");
#endif
  }
  
  /* Mark the DDR as cacheable */
  cache_set_memory_attributes((void*)(0x80000000), 
			      0x7fffffffull, 
			      CACHE_MAR_CACHEABLE | CACHE_MAR_PREFETCHABLE);

  my_rank = cache_get_mm_id();
  if (my_rank == 0) {
    sem_init();
    _register_lock(rtl_lock);
    _register_unlock(rtl_unlock);
  }

  barrier_init();
}

/**
   Exit from the application in a nice fashion.
*/
void
platform_exit()
{
  int my_rank;

  /* Wait for all others to arrive... */
  barrier();

  my_rank = cache_get_mm_id();
  if (my_rank == 0) { 
    exit(1);
  } else {
    abort();
  }
}
