/**
   Simple semaphore routines.
*/

/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "include/sem.h"
#include "include/cache.h"

static uint32_t rtl_lock_depth;

void
sem_init()
{
  sem_registers->rst_run = 0x01;

  while ((sem_registers->rst_run & 0x1) == 0x0);

  rtl_lock_depth = 0;
  cache_l2_wbinv_all(!0);
}

void
sem_lock_spin(unsigned sem)
{
  uint32_t ret;

  do {
    ret = sem_registers->direct[sem];
  } while ((ret & 0xff) != 0);
}

void
sem_unlock(unsigned sem)
{
  sem_registers->direct[sem] = 0x01;
}

int
sem_query(unsigned sem)
{
  uint32_t val;

  val = sem_registers->query[sem];

  if (val == 0x01) {
    return -1;
  } else {
    return (int)((val & 0xff00) >> 8);
  }

}

void 
rtl_lock()
{
  unsigned my_rank;

  my_rank = cache_get_mm_id();
  cache_l2_wbinv_all(!0);

  if (sem_query(0) != my_rank) {
    sem_lock_spin(0);
    rtl_lock_depth = 1;
  } else {
    rtl_lock_depth++;
  }

  cache_l2_wbinv_all(!0);
}

void 
rtl_unlock()
{
  cache_l2_wbinv_all(!0);

  rtl_lock_depth--;
  if (rtl_lock_depth == 0) {
    sem_unlock(0);
  }

  cache_l2_wbinv_all(!0);
}
