/*

Copyright (c) 2012 Kungliga Tekniska Högskolan 
(Royal Institute of Technology, Stockholm, Sweden).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

*/

#include "uart.h"
#include "uart_regs.h"

static const uint32_t uart_div_115200 = 90; /* Magic 115200 baud. */
static const unsigned uart_tx_blk_size = 16; /* 16 byte tx FIFO. */

void 
uart_init()
{
  /* Make sure that the UART is not in reset. */
  uart_registers->pwremu_mgmt = 0x7fff;

  /* Set up frame format to 8N1. disable DLAB access via RBR */
  uart_registers->lcr = 0x03;

  /* 16x oversampling mode. */
  uart_registers->mdr = 0x00;

  /* Set up the baud rate divisor latches. */
  uart_registers->dll = uart_div_115200 & 0xff;
  uart_registers->dlh = (uart_div_115200 >> 8) & 0xff;

  /* Clear and enable the FIFOs, polling mode. */
  uart_registers->iir_fcr = 0xcf;

  /* Disable all interrupt generation. */
  uart_registers->ier = 0x00;
  
  /* Disable autoflow handshake, no loop-back. */
  uart_registers->mcr = 0x00;
}


void 
uart_puts(const char *s)
{
  unsigned cnt=0;
  const char *p=s;

  while (*p != '\0') {
    if (cnt == 0) {
      while ((uart_registers->lsr & 0x20) != 0x20) ;
      cnt = uart_tx_blk_size;
    }
    
    uart_registers->rbr_thr = (uint32_t)(*p);

    ++p;
    --cnt;
  }
}

